/*
 * DatagramLogListener.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_IO_PKG;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * A log listener that tracks the log level and can reply to StatusMessage
 * queries about the level.  This is only really useful for getting a quick
 * summary of the status of a node.
 */
public class DatagramLogListener
	implements LogListener
{
	/**
	 * The localhost InetAddress
	 */
	private static InetAddress localAddr;

	/**
	 * The level we reply to queries with
	 */
	private int currentLevel = 0;

	static
	{
		try
		{
			localAddr = InetAddress.getLocalHost();
		}
		catch(UnknownHostException e)
		{
		}
	}

	/**
	 * The server that replies to queries about the status of the node
	 */
	private class SocketListener
		implements Runnable
	{
		/**
		 * Listener socket
		 */
		private DatagramSocket ds;
		
		/**
		 * Reply message
		 */
		private StatusMessage outMessage = new StatusMessage();

		/**
		 * Incoming message
		 */
		private StatusMessage inMessage = new StatusMessage();

		/**
		 * Construct a listener for a certain port
		 */
		public SocketListener(int port)
			throws Exception
		{
			this.ds = new DatagramSocket(port);
		}

		/**
		 * Process an incoming packet
		 */
		public void processPacket()
		{
			if( inMessage.getDatagramPacket().getLength() <
			    inMessage.getValueOffset() )
			{
				return;
			}
			switch( inMessage.getMethodID() )
			{
			case StatusMessage.SMM_QUERY:
				/* Build the reply */
				outMessage.setMethodID(StatusMessage.
						       SMM_QUERY);
				outMessage.setFlags(StatusMessage.SMF_REPLY);
				StatusMessage.encodeInt(
					outMessage.getValue(),
					outMessage.getValueOffset(),
					currentLevel);
				outMessage.setValueSize(4);
				outMessage.setAddress(inMessage.getAddress());
				outMessage.setPort(inMessage.getPort());
				try
				{
					/* Send it back to the querier */
					this.ds.send(outMessage.
						     getDatagramPacket());
				}
				catch(IOException e)
				{
				}
				break;
			}
		}
		
		public void run()
		{
			/**
			 * Loop forever processing requests
			 */
			while( true )
			{
				try
				{
					this.ds.receive(this.inMessage.
							getDatagramPacket());
					this.processPacket();
				}
				catch(IOException ioe)
				{
				}
			}
		}
	}

	/**
	 * Start a listener on the given port
	 */
	public void start(int port)
		throws Exception
	{
		Thread th;

		th = new Thread(new SocketListener(port));
		th.start();
	}

	/**
	 * Receive the log message, we just record the level for now.
	 */
	public void logMsg(int level, String msg)
	{
		/*
		 * XXX This sucks, we should be more intelligent about the
		 * level
		 */
		this.currentLevel = level;
	}
}
