/*
 * Assert.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

IMPORT_DEBUG

/**
 * Assert package.
 * <p>
 * XXX Currently this is public for my convenience.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class Assert 
{
#ifdef DEBUG
	/**
	 * If an assert fails, this routine is called to print the
	 * file and line number information, along with the string
	 * representation of the failed expression, and a backtrace
	 * to the failure.
	 *
	 * @param str the string representation of the expression
	 * @param file the name of the file in which the assert failed
	 * @param line the line number on which the assert failed in <code>file</code>
	 */
	private static void assertFailure(String typemsg, String str, String file, int line)
	{
		NATIVE_DEBUG.beginDebugOutput();
		Debug.printLocation(file, line);
		NATIVE_DEBUG.print(": " +typemsg+ " failed: ");
		NATIVE_DEBUG.println(str);
		NATIVE_DEBUG.endDebugOutput();
		
		Debug.printStackTrace();

		NATIVE_DEBUG.exit();
	}
	
	/**
	 * Check that the boolean expression is true.  Calls assertFailure()
	 * if not.
	 *
	 * @param expr the expression to check
	 * @param exprStr the string representation of the expression
	 * @param file the name of the file the assert is in
	 * @param line the line number of the file the assert is in
	 * @see #assertFailure(String,String,String,String)
	 */
	public static void assertCheck(boolean expr, String exprStr, String file, int line) 
	{
		if (!expr) 
		{
			assertFailure("assert", exprStr, file, line);
		}
	}

	/**
	 * Check that the boolean expression is true.  Calls assertFailure()
	 * if not.
	 *
	 * @param expr the expression to check
	 * @param exprStr the string representation of the expression
	 * @param file the name of the file the require is in
	 * @param line the line number of the file the require is in
	 * @see #assertFailure(String,String,String,String)
	 */
	public static void requireCheck(boolean expr, String exprStr, String file, int line) 
	{
		if (!expr) 
		{
			assertFailure("require", exprStr, file, line);
		}
	}

	/**
	 * Check that the boolean expression is true.  Calls assertFailure()
	 * if not.
	 *
	 * @param expr the expression to check
	 * @param exprStr the string representation of the expression
	 * @param file the name of the file the ensure is in
	 * @param line the line number of the file the ensure is in
	 * @see #assertFailure(String,String,String,String)
	 */
	public static void ensureCheck(boolean expr, String exprStr, String file, int line) 
	{
		if (!expr) 
		{
			assertFailure("ensure", exprStr, file, line);
		}
	}

	/**
	 * Check that the given object reference is non-null. Calls assertFailure() if the reference
	 * is null.
	 *
	 * @param expr the expression to check
	 * @param exprStr the string representation of the expression
	 * @param file the name of the file the assert is in
	 * @param line the line number of the file the assert is in
	 * @see #assertFailure(String,String,String,String)
	 */
	public static void assertCheck(Object ref, String exprStr, String file, int line) 
	{
		if (ref == null) 
		{
			assertFailure("assert", exprStr, file, line);
		}
	}

	/**
	 * Check that the given object reference is non-null. Calls
	 * assertFailure() if the reference is null.
	 *
	 * @param expr the expression to check
	 * @param exprStr the string representation of the expression
	 * @param file the name of the file the assert is in
	 * @param line the line number of the file the assert is in
	 * @see #assertFailure(String,String,String,String)
	 */
	public static void requireCheck(Object ref, String exprStr, String file, int line) 
	{
		if (ref == null) 
		{
			assertFailure("require", exprStr, file, line);
		}
	}

	/**
	 * Check that the given object reference is non-null. Calls assertFailure() if the reference
	 * is null.
	 *
	 * @param expr the expression to check
	 * @param exprStr the string representation of the expression
	 * @param file the name of the file the assert is in
	 * @param line the line number of the file the assert is in
	 * @see #assertFailure(String,String,String,String)
	 */
	public static void ensureCheck(Object ref, String exprStr, String file, int line) 
	{
		if (ref == null) 
		{
			assertFailure("ensure", exprStr, file, line);
		}
	}

	/**
	 * Assert the given object is locked.
	 */
	public static void assertLockedCheck(Object obj, String exprStr, String file, int line)
	{
		if (! NATIVE_DEBUG.checkLocked(obj))
		{
			assertFailure("assertLocked", exprStr, file, line);
		}
	}
	
#endif
}
