/*
 * BufferHandle.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

/**
 * A Java wrapper around the native buffer handle object.  The Java
 * wrapper protects the NodeOS buffer handle from errant or malicious
 * Java code.
 * <ul>
 *  <li>This layer does the run-time bounds checks on reads and
 *     writes.</li>
 *  <li>Provides a layer of indirection so that the BufferHandle
 *     can be revoked (for example when passed to an InChannel)</li>
 * </ul>
 * There is a one-to-one correspondence between a Java BufferHandle
 * and the underlying NodeOS buffer handle (pbuf).
 *
 * <p>
 * A BufferHandle is a process's handle to a buffer.  A BufferHandle
 * object is seen by only one process--it cannot be shared
 * between processes.  However, the underlying buffer object may be
 * visible in more than one process.  The underlying buffer is never
 * directly accessible to a process, it can only be accessed
 * through the BufferHandle.
 * <p>
 * The "offset" and "length" associated with a buffer handle are
 * immutable once a buffer is associated with the handle.  They
 * can thus be used to provide safe views onto subsequences of
 * a buffer.
 * <p>
 * A process is charged for the memory occupied by all buffers reachable
 * from the process's BufferHandle objects.  To avoid paying for a buffer
 * which is no longer needed, a process can explicitly <code>reset()</code> a
 * BufferHandle object, which tells Janos that the process no longer
 * needs the buffer data associated with the BufferHandle.
 * An access to a BufferHandle after it has been recycled behaves
 * unpredictably: it may continue to access the old buffer, it
 * may access another one of the processes's buffers, or it may
 * raise a NullPointerException.  However, such an access will never
 * read another process's buffer.
 * <p>
 * Currently, all access methods in BufferHandle are synchronized.
 * This is inefficient.  The virtual machine can eliminate these locks
 * in a section of code by incrementing a per-BufferHandle reference
 * count before the code section and decrementing the reference count
 * after the code section (and postponing all resets until the
 * reference count is zero).  Unfortunately, this optimization cannot
 * be expressed at the Java source level (without using
 * synchronization).
 * <p>
 * Since a BufferHandle is visible only to one process, and a process
 * can only run on one processor at a time, simple preemption points
 * can eliminate the per-access synchronization.
 * <p>
 * Buffers are mutable.
 * <p>
 * This class is not final so that you can subclass it an add things like
 * buffer aggregates on top.  XXX need to be sure NATIVE_BUFFERHANDLE
 * methods are not exposed...
 * <p>
 * XXX Perhaps BufferHandle shouldn't be doing any run-time checks?  That
 * should be the responsibility of the EE, right?
 * <p>
 * XXX WARNING DO NOT ADD ANY INSTANCE FIELDS TO THIS CLASS WITHOUT
 * CHANGING THE GLUE CODE IN moabthread.c.  THAT CODE ASSUMES THAT
 * "owner" IS THE FIRST AND ONLY FIELD IN THIS CLASS.  THAT IS A 
 * STRONG ASSUMPTION TO MAKE.  ALSO ASSUMES THE VALUE '2' IS FOR
 * THE USER_OWNER.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class BufferHandle
	extends NATIVE_BUFFERHANDLE
{
	/**
	 * Constant used for the <code>owner</code> field.
	 * Indicates that the system has control over this buffer.
	 *
	 * @see #owner
	 */
	private static final int SYSTEM_OWNER = 1;

	/**
	 * Constant used for the <code>owner</code> field.
	 * Indicates that the user has control over this buffer.
	 *
	 * @see #owner
	 */
	private static final int USER_OWNER = 2;

#ifdef DEBUG
	/**
	 * A bit of a debugging hack.  All the "native" classes
	 * contain a bit of debugging magic that asserts that this
	 * field is what they expect it to be.  Look at
	 * java/Defines.cpp.
	 */
	PACKAGE static final int NativeNodeOSID = NATIVE_NODEOS;
#endif

	/**
	 * A BufferHandle is owned by the user when it is created.
	 * It becomes owned by the system when passed to InChannel.addBuffer().
	 * Ownership is restored to the user when the buffer is passed
	 * to a recieve callback.
	 *
	 * XXX this field may not be necessary (except for debugging Moab).
	 * (May be required for reset(),getInt() sequence checks...
	 *
	 * XXX WARNING DO NOT ADD ANY INSTANCE FIELDS TO THIS CLASS WITHOUT
	 * CHANGING THE GLUE CODE IN moabthread.c.  THAT CODE ASSUMES THAT
	 * "owner" IS THE FIRST AND ONLY FIELD IN THIS CLASS.  THAT IS A 
	 * STRONG ASSUMPTION TO MAKE.
	 *
	 * @see InChannel#addBuffer()
	 * @see DeliverFunc()
	 */
	private int owner;


	/**
	 * Construct a new BufferHandle.  Constructs a buffer handle that
	 * points to a newly allocated, maximum size buffer.  Offset
	 * is zero.
	 */
	public static BufferHandle create() 
	{
		return create(null, 0, NATIVE_BUFFERHANDLE.getMaxPacketLen());
	}

	/**
	 * Constructs a new buffer handle with the given offset and "length".
	 * (The underlying buffer may have a greater actual offset, and may
	 * be larger than offset+length bytes.)
	 */
	public static BufferHandle create(int offset, int length)
	{
		return create(null, offset, length);
	}
	
	/**
	 * Construct a new BufferHandle that points to the same data
	 * as the given handle.  (Shallow copy of other bufferhandle).
	 *
	 * @see #cloneHandle(BufferHandle)
	 * @param bh the BufferHandle to copy from
	 */
	public static BufferHandle create(BufferHandle other)
	{
		return create(other, other.getOffset(), other.getLength());
	}
	
	/**
	 * Construct a new BufferHandle that points to the same data
	 * as the given handle.  (Shallow copy of other bufferhandle).
	 * The new buffer handle's offset and length are set to the
	 * provided values.
	 *
	 * @see #cloneHandle(BufferHandle)
	 * @param other the buffer handle to "copy" from (or null).
	 * @param offset the new offset this bufferhandle will have.
	 * @param length the new length this bufferhandle will have.
	 */
	public static BufferHandle create(BufferHandle other, int offset, int length)
	{
		final int maxplen = NATIVE_BUFFERHANDLE.getMaxPacketLen();
		
		if (length < 0)
			throw new NegativeArraySizeException("Negative length in BufferHandle.create()");

		if ((offset < 0)
		    || (offset > maxplen)
		    || (length > maxplen)
		    || (offset+length > maxplen))
			// XXX this is a silly error to throw, no?
			throw new ArrayIndexOutOfBoundsException("Bogon parameters passed to BufferHandle.create()");

		if (other != null) 
		{
			require(other.getOffset() <= maxplen);
			require(other.getLength() <= maxplen);
			require(other.getLength()+other.getOffset() < maxplen);

			if ((other.getOffset() + offset + length > maxplen)
			    || (offset + length > other.getLength()))
				throw new ArrayIndexOutOfBoundsException("exceeded the limits of source buffer.");
		}

		return new BufferHandle(other, offset, length);
	}
	
	/**
	 * Construct a new buffer handle.  If bh is null, create a new
	 * underlying buffer, otherwise point the new buffer handle
	 * at the same buffer as 'other'.
	 *
	 * @param other <code>null</code> or a buffer handle to clone
	 * @param offset initial offset
	 * @param length length of data
	 */
	private BufferHandle(BufferHandle other, int offset, int length)
	{
		super(other==null?null:other.getNativeBufH(),
		      offset, length);
		this.owner = USER_OWNER;
	}

	/**
	 * Release the buffer held by this BufferHandle.  Subsequent
	 * accesses to the buffer must fail.
	 */
	public final synchronized void reset()
	{
		if (this.owner != USER_OWNER)
			throw new Error("You don't own this buffer."); // XXX BufferOwnerError

		this.owner = SYSTEM_OWNER;
		super.reset();
	}

	/**
	 * Point this BufferHandle to a new buffer.
	 *
	 * @allocation Allocates a new underlying buffer of at least
	 * length bytes
	 */
	public final synchronized void newBuffer(int offset, int length)
		throws ArrayIndexOutOfBoundsException,
		NegativeArraySizeException
	{
		if (length < 0)
			throw new NegativeArraySizeException("length is negative");
		
		if ((offset < 0)
		    || (offset > NATIVE_BUFFERHANDLE.getMaxPacketLen())
		    || (length > NATIVE_BUFFERHANDLE.getMaxPacketLen())
		    || (offset + length > NATIVE_BUFFERHANDLE.getMaxPacketLen()))
			throw new ArrayIndexOutOfBoundsException();
		
		/*
		 * Technically, its okay for a buffer queued on an InChannel
		 * to be newBuffer'd.  Its stupid, yes, but it won't violate
		 * the corectness of the system.
		 */
		if (this.owner != USER_OWNER)
			throw new Error("You gave this buffer away already");

		super.newBuffer(offset, length);
	}

	/**
	 * Set this BufferHandle's buffer to point to some
	 * sub-sequence of src's buffer.
	 *
	 * @allocation allocates exception objects if there is a problem.
	 */
	public final void cloneHandle(BufferHandle src, int fromOffset, int fromLength)
		throws ArrayIndexOutOfBoundsException,
		NegativeArraySizeException
	{
		if (src == null)
			throw new NullPointerException("source buffer in cloneHandle() is null");

		if (fromLength < 0)
			throw new NegativeArraySizeException();

		if ((fromOffset < 0)
		    || (fromOffset > src.getLength()) // might int-overflow with '+'
		    || (fromLength > src.getLength()) // might int-overflow with '+'
		    || (fromOffset + fromLength > src.getLength()))
			throw new ArrayIndexOutOfBoundsException(/*XXX*/);

		/*
		 * Technically, its okay for a buffer queued on an
		 * In/Out Channel to be cloneHandle'd.  Its stupid, yes,
		 * but it won't violate the corectness of the system.
		 */
		if (src.owner != USER_OWNER)
			throw new Error("Source buffer not owned by user.");

		// Prevent access/changes to this BufferHandle while
		// copying the src buffer

		synchronized(this)
		{
			super.cloneHandle(src.getNativeBufH(),
					  fromOffset, fromLength);
		}
	}

	/**
	 * Set this BufferHandle's buffer to point to src's buffer.
	 */
	public final void cloneHandle(BufferHandle src)
	{
		// Argument checking is done in called cloneHandle()

		cloneHandle(src.getNativeBufH(),
			    0, src.getLength());
	}

	/**
	 * Get the valid data length of the underlying NodeOS pbuf.
	 */
	public final int getLength()
	{
		// No need to check this.user

		return super.getLength();
	}

	public final synchronized byte getByte(int index)
		throws ArrayIndexOutOfBoundsException
	{
		if ((index < 0)
		    || (index >= super.getLength()))
			throw new ArrayIndexOutOfBoundsException();

		/*
		 * Something in user-land is broken if they're getting
		 * bytes out of a InChannel-queued BufferHandle.  But,
		 * it doesn't violate the NodeOS to do so.
		 */
#ifdef DEBUG
		if (this.owner != USER_OWNER)
			throw new Error("User does not own this buffer.");
#endif

		return super.getByte(index);
	}

	public final synchronized void setByte(int index, byte val)
		throws ArrayIndexOutOfBoundsException
	{
		if ((index < 0)
		    || (index >= super.getLength()))
			throw new ArrayIndexOutOfBoundsException();
		/*
		 * Something in user-land is broken if they're getting
		 * bytes out of a InChannel-queued BufferHandle.  But,
		 * it doesn't violate the NodeOS to do so.
		 */
#ifdef DEBUG
		if (this.owner != USER_OWNER)
			throw new Error("User does not own this buffer.");
#endif

		super.setByte(index, val);
	}

	public final synchronized short getShort(int index)
		throws ArrayIndexOutOfBoundsException
	{
		if ((index < 0)
		    || (index + 2 > super.getLength()))
			throw new ArrayIndexOutOfBoundsException();

		/*
		 * Something in user-land is broken if they're getting
		 * bytes out of a InChannel-queued BufferHandle.  But,
		 * it doesn't violate the NodeOS to do so.
		 *
		 * XXX that's not right... it is bad, as the buffer is
		 * shared.  Not bad if buffer is detached from handle
		 * by Moab...
		 */
#ifdef DEBUG
		if (this.owner != USER_OWNER)
			throw new Error("User does not own this buffer.");
#endif

		return super.getShort(index);
	}

	public final synchronized void setShort(int index, short val)
			throws ArrayIndexOutOfBoundsException
	{
		if((index < 0) ||
		   (index >= super.getLength()-1))
			throw new ArrayIndexOutOfBoundsException();

		/*
		 * Something in user-land is broken if they're getting
		 * bytes out of a InChannel-queued BufferHandle.  But,
		 * it doesn't violate the NodeOS to do so.
		 */
#ifdef DEBUG
		if (this.owner != USER_OWNER)
			throw new Error("User does not own this buffer.");
#endif

		super.setShort(index, val);
	}

	public final synchronized int getInt(int index)
		throws ArrayIndexOutOfBoundsException
	{
		if ((index < 0)
		    || (index + 4 > super.getLength()))
			throw new ArrayIndexOutOfBoundsException();

		/*
		 * Something in user-land is broken if they're getting
		 * bytes out of a InChannel-queued BufferHandle.  But,
		 * it doesn't violate the NodeOS to do so.
		 */
#ifdef DEBUG
		if (this.owner != USER_OWNER)
			throw new Error("User does not own this buffer.");
#endif

		return super.getInt(index);
	}

	public final synchronized void setInt(int index, int val)
			throws ArrayIndexOutOfBoundsException
	{
		if((index < 0) ||
		   (index >= super.getLength()-3))
			throw new ArrayIndexOutOfBoundsException();

		/*
		 * Something in user-land is broken if they're getting
		 * bytes out of a InChannel-queued BufferHandle.  But,
		 * it doesn't violate the NodeOS to do so.
		 */
#ifdef DEBUG
		if (this.owner != USER_OWNER)
			throw new Error("User does not own this buffer.");
#endif

		super.setInt(index, val);
	}

	public final void getBytes(byte[] dest, int destPosition)
	{
		getBytes(0, dest, destPosition, getLength());
	}

	public final synchronized void getBytes(int srcPosition, byte[] dest,
			int destPosition, int destLength)
			throws ArrayIndexOutOfBoundsException
	{
		if((srcPosition < 0)
		   || (srcPosition + destLength > super.getLength())
		   || (destPosition < 0)
		   || (destPosition + destLength > dest.length))
			throw new ArrayIndexOutOfBoundsException();

		/*
		 * Something in user-land is broken if they're getting
		 * bytes out of a InChannel-queued BufferHandle.  But,
		 * it doesn't violate the NodeOS to do so.
		 */
#ifdef DEBUG
		if (this.owner != USER_OWNER)
			throw new Error("User does not own this buffer.");
#endif

		super.getBytes(srcPosition, dest, destPosition, destLength);
	}

	public final synchronized void setBytes(int destPosition, byte[] src,
			int srcPosition, int srcLength)
			throws ArrayIndexOutOfBoundsException
	{
		if((destPosition < 0)
		   || (destPosition + srcLength > super.getLength()))
			throw new ArrayIndexOutOfBoundsException();

		// XXX if (srcLength > src.length!) ... 

		/*
		 * Something in user-land is broken if they're getting
		 * bytes out of a InChannel-queued BufferHandle.  But,
		 * it doesn't violate the NodeOS to do so.
		 */
#ifdef DEBUG
		if (this.owner != USER_OWNER)
			throw new Error("User does not own this buffer.");
#endif

		super.setBytes(destPosition, src, srcPosition, srcLength);
	}

	public final synchronized void getInts(int srcPosition, int[] dest,
			int destPosition, int destLength)
		throws ArrayIndexOutOfBoundsException
	{
		if((srcPosition < 0)
		   || (srcPosition + destLength*4 > super.getLength())
		   || (destPosition < 0)
		   || (destPosition + destLength > dest.length))
			throw new ArrayIndexOutOfBoundsException();

		/*
		 * Something in user-land is broken if they're getting
		 * bytes out of a InChannel-queued BufferHandle.  But,
		 * it doesn't violate the NodeOS to do so.
		 */
#ifdef DEBUG
		if (this.owner != USER_OWNER)
			throw new Error("User does not own this buffer.");
#endif

		super.getInts(srcPosition, dest, destPosition, destLength);
	}

	public final synchronized void setInts(int destPosition, int[] src,
			int srcPosition, int srcLength)
		throws ArrayIndexOutOfBoundsException
	{
		if((destPosition < 0)
		   || (destPosition + srcLength*4 > super.getLength()))
			throw new ArrayIndexOutOfBoundsException();

		// XXX if (srcLength > src.length!) ... 

		/*
		 * Something in user-land is broken if they're getting
		 * bytes out of a InChannel-queued BufferHandle.  But,
		 * it doesn't violate the NodeOS to do so.
		 */
#ifdef DEBUG
		if (this.owner != USER_OWNER)
			throw new Error("User does not own this buffer.");
#endif

		super.setInts(destPosition, src, srcPosition, srcLength);
	}

	/**
	 * Compute an IP style checksum over the given segment of the buffer.
	 *
	 * @param offset Offset into the buffer to start computing the checksum
	 * @param count The number of bytes the checksum should cover.
	 * @return The checksum value.
	 */
	public final synchronized short computeChecksum(int offset, int count)
	{
		if( (offset < 0) ||
		    (offset >= this.getLength()) ||
		    ((offset + count) > this.getLength()) )
		{
			throw new ArrayIndexOutOfBoundsException();
		}

		return super.computeChecksum(offset, count);
	}

	/**
	 * Compare the given sequence of bytes to a sequence in this buffer.
	 * Returns true if the sequences are identical, false if not.
	 * <p>
	 * XXX a more generic version that allows sub-sequences of 'bytes' might be
	 * handy.
	 *
	 * @param offset Offset into the buffer to start comparison
	 * @param bytes The byte sequence to compare with.
	 */
	public final synchronized boolean compareBytes(int offset, byte[] bytes)
		throws ArrayIndexOutOfBoundsException
	{
		if ((offset < 0)
		    || (offset + bytes.length > super.getLength()))
			throw new ArrayIndexOutOfBoundsException();

		if (bytes == null)
			throw new NullPointerException("null bytes array passed to compareBytes()");

		return super.compareBytes(offset, bytes, 0, bytes.length);
	}

	/*
	 * Get the native buffer handle associated with this
	 * BufferHandle.
	 *
	 * @return the native buffer handle
	 */
	PACKAGE synchronized NATIVE_BUFFERHANDLE getNativeBufH()
	{
		return this;
	}

	/**
	 * Change ownership of this BufferHandle from user to system.
	 */
	PACKAGE void systemOwns()
	{
		require(this.owner == BufferHandle.USER_OWNER);
		this.owner = BufferHandle.SYSTEM_OWNER;
	}

	/**
	 * Change ownership of this BufferHandle from system to user.
	 */
	PACKAGE void userOwns()
	{
		require(this.owner == BufferHandle.SYSTEM_OWNER);
		this.owner = BufferHandle.USER_OWNER;
	}

	public String toString()
	{
		return super.toString();
	}
}

