/*
 * CommSpace.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.lang.ArrayIndexOutOfBoundsException;

import JVM_SYS_PKG.Exportable;
import JVM_SYS_PKG.TemporaryExport;

/**
 * Variation on an array, with some tuple-space like properties.
 * Objects are not shared between flows, they are copied on read.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class CommSpace
	extends Exportable
	implements TemporaryExport
{
	/**
	 * The table mapping indicies to elements.
	 */
	private final CommSpaceElement[] table;

	/**
	 * @param size The size of CommSpace to construct.
	 */
	PACKAGE CommSpace(int size)
	{
		this.table = new CommSpaceElement[size];
	}

	/**
	 * @return The size of the CommSpace.
	 */
	PACKAGE final int size()
	{
		return this.table.length;
	}

	/**
	 * @param index Index to verify that it is within the bounds of the
	 * space.
	 * @exception ArrayIndexOutOfBoundsException when the index is not
	 * valid.
	 */
	private final void checkIndex(int index)
	{
		if ((index >= this.table.length)
		    || (index < 0))
			throw new ArrayIndexOutOfBoundsException("CommSpace table only has "
								 +size()+ " elements");
	}

	/**
	 * Write the given element at the given index.
	 */
	PACKAGE void write(int index, CommSpaceElement elem)
		throws ArrayIndexOutOfBoundsException
	{
		checkIndex(index);
		CommSpaceElement elemCopy = CommSpaceElement.copy(elem);

		TRACE(Trace.NODEOS_COMMSPACE, "write(idx="+index+",elem="+elem+"; table="+this+")");
		synchronized(this.table)
		{
			this.table[index] = elemCopy;
			notifyOnIndex(index);
		}
	}

	PACKAGE CommSpaceElement take(int index)
		throws ArrayIndexOutOfBoundsException
	{
		CommSpaceElement e;

		TRACE(Trace.NODEOS_COMMSPACE, "take(idx="+index+"; table="+this+")");
		checkIndex(index);
		synchronized(this.table)
		{
			e = this.table[index];
			this.table[index] = null;
		}
		return e;
	}

	PACKAGE CommSpaceElement waitToTake(int index)
		throws ArrayIndexOutOfBoundsException, InterruptedException
	{
		CommSpaceElement e;

		TRACE(Trace.NODEOS_COMMSPACE, "waitToTake(idx="+index+"; table="+this+")");

		checkIndex(index);
		synchronized(this.table)
		{
			e = this.table[index];
			while (e == null)
			{
				waitOnIndex(index);
				e = this.table[index];
			}
			this.table[index] = null;
		}
		return e;
	}

	PACKAGE CommSpaceElement read(int index)
		throws ArrayIndexOutOfBoundsException
	{
		CommSpaceElement e;

		TRACE(Trace.NODEOS_COMMSPACE, "read(idx="+index+"; table="+this+")");

		checkIndex(index);
		synchronized(this.table)
		{
			e = this.table[index];
		}
		return e;
	}

	PACKAGE CommSpaceElement swap(int index, CommSpaceElement elem)
	{
		CommSpaceElement e;
		
		TRACE(Trace.NODEOS_COMMSPACE, "swap(idx="+index+",elem="+elem+"; table="+this+")");

		checkIndex(index);
		synchronized(this.table)
		{
			e = this.table[index];
			this.table[index] = CommSpaceElement.copy(elem);
			notifyOnIndex(index);
		}
		return e;
	}

	// Assume there won't be a lot of contention/waiting in a
	// single CommSpace, and just wait/signal on the whole table,
	// not individual elements.

	/**
	 * Notify potential interested parties about a change at the
	 * given index.
	 */
	private void notifyOnIndex(int index)
	{
		// table.assertSynchronizeLocked()
		TRACE(Trace.NODEOS_COMMSPACE, " notifyOnIndex(idx="+index+"; table="+this+")");
		this.table.notifyAll();
	}

	/**
	 * Wait for something to change in the slot at the
	 * given index.
	 */
	private void waitOnIndex(int index)
		throws InterruptedException
	{
		// table.assertSynchronizeLocked()
		TRACE(Trace.NODEOS_COMMSPACE, " waitOnIndex(idx="+index+"; table="+this+")");
		this.table.wait();
		TRACE(Trace.NODEOS_COMMSPACE, " WOKEN OnIndex(idx="+index+"; table="+this+")");
	}

	public String toString()
	{
		return "CommSpace[t=@"+Integer.toHexString(this.table.hashCode())+"]";
	}
}
