/*
 * CommSpaceHandle.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.lang.ArrayIndexOutOfBoundsException;

import JVM_SYS_PKG.Importable;
import JVM_SYS_PKG.Exportable;
import JVM_SYS_PKG.CarrierVisitor;
import JVM_SYS_PKG.ExceptionInVisit;
import JVM_SYS_PKG.DeadTeamException;
import JVM_SYS_PKG.ThrowInVisitException;

/**
 * A handle on a CommSpace. A CommSpace is an inter-process communication
 * facility based lightly on Linda's "tuples".  The space consists of
 * an array of CommSpaceElements.  Elements are copied out on read.  A
 * thread may write, read, swap, "wait" or "take" any element.
 * <p>
 * XXX Semantics during a failed take() are probably wrong.
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 * @see CommSpace
 */
public final class CommSpaceHandle
	extends Importable
{
	/**
	 * Create a new CommSpace of the given size and return a handle to
	 * it.
	 *
	 * @param size The size of the CommSpace.
	 * @return The handle to the CommSpace.
	 */
	public static CommSpaceHandle newCommSpace(int size)
	{
		return new CommSpaceHandle(new CommSpace(size));
	}

	/**
	 * Create a CommSpace and attach it to the given handle.
	 *
	 * @param csh The CommSpaceHandle to attach to the newly created
	 * CommSpace.
	 * @param size The size of the CommSpace.
	 * @return The handle that was passed in.
	 */
	public static CommSpaceHandle newCommSpace(CommSpaceHandle csh,
						   int size)
	{
		assert(csh); // XXX
		assert(csh.empty()); // XXX

		new CommSpace(size).subscribe(csh);
		return csh;
	}

	/**
	 * Create an empty handle.
	 *
	 * @see #newCommSpace(CommSpaceHandle,int)
	 */
	public CommSpaceHandle()
	{
	}

	/**
	 * Copy constructor.  Only used when passing a
	 * csh to a child flow.
	 *
	 * @param other The handle to copy.
	 */
	PACKAGE CommSpaceHandle(CommSpaceHandle other)
	{
		require(other);
		
		this.copy(other);
	}

	/**
	 * Construct a handle that refers to the given CommSpace.
	 *
	 * @param commSpace The CommSpace this handle should refer too.
	 */
	PACKAGE CommSpaceHandle(CommSpace commSpace)
	{
		commSpace.subscribe(this);
	}

	/**
	 * XXX Without this being public there is no way to tell if this
	 * thing actually points to a commSpace.
	 */
	PACKAGE boolean empty()
	{
		return this.get() == null;
	}

	/**
	 * Return the number of slots in the CommSpace pointed to by
	 * this handle.
	 */
	public int size()
	{
		/* XXX this.get() could be null */
		return ((CommSpace)this.get()).size();
	}

	/**
	 * Resize the CommSpace pointed to by this handle to contain
	 * the given number of slots.
	 * <p>
	 * XXX what happens when resizing to smaller than current and
	 * dropping items?
	 */
	public int resize(int newsize)
	{
		NOTIMPLEMENTED("resize()");

		// XXX only doable by allocator of underlying CommSpace object

		// watch out for shrinkage.
	}

	/**
	 * Visit the CommSpace and run the given code.  If code.value is
	 * non-null then it is copied and returned.
	 *
	 * @param code The code to run in the CommSpace.
	 * @return A copy of the value captured in code.value.
	 */
	private CommSpaceElement visitCommSpace(final CommSpace cs,
						final CarrierVisitor code)
		throws ArrayIndexOutOfBoundsException
	{
		CommSpaceElement retval = null;
		
		try
		{
			cs.pushOwnerActivation().visit(code);
			if( code.value != null )
			{
				CommSpaceElement cse = (CommSpaceElement)code.
					value;

				retval = (CommSpaceElement)cse.clone();
			}
		}
		catch(DeadTeamException dte)
		{
		}
		catch(ExceptionInVisit eiv)
		{
			if( eiv.getException() instanceof
			    ArrayIndexOutOfBoundsException )
			{
				cs.popOwnerActivation();
				throw (ArrayIndexOutOfBoundsException)
					eiv.getException();
			}
		}
		catch(ThrowInVisitException eiv)
		{
			/* Should we do something here? */
		}
		cs.popOwnerActivation();
		return retval;
	}

	/**
	 * Write the given element to the given index.  Overwrites
	 * whatever is already there.
	 *
	 * @param index The index in the CommSpace to write into.
	 * @param elem The element to copy into the CommSpace.
	 */
	public void write(final int index, final CommSpaceElement elem)
		throws ArrayIndexOutOfBoundsException
	{
		final CommSpace cs = (CommSpace)this.get();

		this.visitCommSpace(cs, new CarrierVisitor() {
			public void run()
				throws Exception
			{
				cs.write(index, elem);
			}
		});
	}
	
	/**
	 * Take an element from the given index and put a null in its place.
	 * The read and replace are atomic with respect to other take()s at the
	 * given index. XXX semantically equivalent to
	 * swap(<var>index</var>, <strong>null</strong>)
	 *
	 * @param index The index in the CommSpace that we should take.
	 * @see #swap(int,CommSpaceElement)
	 * @return null or the CommSpaceElement that was in the given index.
	 */
	public CommSpaceElement take(final int index)
		throws ArrayIndexOutOfBoundsException
	{
		final CommSpace cs = (CommSpace)this.get();

		return this.visitCommSpace(cs, new CarrierVisitor() {
			public void run()
				throws Exception
			{
				this.value = cs.take(index);
			}
		});
	}
	
	/**
	 * Wait for an element to be placed into the given index.  Then
	 * atomically take the element and replace with null.
	 *
	 * @param index The index in the CommSpace that we should take.
	 * @return the CommSpaceElement that was in the given index.
	 */
	public CommSpaceElement waitToTake(final int index)
		throws InterruptedException, ArrayIndexOutOfBoundsException
	{
		final CommSpace cs = (CommSpace)this.get();
		CommSpaceElement retval;
		
		retval = this.visitCommSpace(cs, new CarrierVisitor() {
			public void run()
				throws Exception
			{
				this.value = cs.waitToTake(index);
			}
		});
		if( retval == null )
			throw new InterruptedException();
		return retval;
	}
	
	/**
	 * Read the element at the given index.
	 *
	 * @param index The index in the CommSpace that we should take.
	 * @return Null or the CommSpaceElement that was in the given index.
	 */
	public CommSpaceElement read(final int index)
		throws ArrayIndexOutOfBoundsException
	{
		final CommSpace cs = (CommSpace)this.get();

		return this.visitCommSpace(cs, new CarrierVisitor() {
			public void run()
				throws Exception
			{
				this.value = cs.read(index);
			}
		});
	}
	
	/**
	 * Swap an element with the one currently in the given index.
	 *
	 * @param index The index in the CommSpace that we should take.
	 * @param elem The element to swap into the index.
	 * @return Null or the CommSpaceElement that was in the given index.
	 */
	public CommSpaceElement swap(final int index,
				     final CommSpaceElement elem)
		throws ArrayIndexOutOfBoundsException
	{
		final CommSpace cs = (CommSpace)this.get();

		return this.visitCommSpace(cs, new CarrierVisitor() {
			public void run()
				throws Exception
			{
				this.value = cs.swap(index, elem);
			}
		});
	}
}
