/*
 * Debug.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

/**
 * The package-internal debugging infrastructure.
 * <p>
 * The majority of routines in this file deal with printing directly
 * to stdout/stderr through VM facilities.  See the corresponding
 * native layer for the details on where that output is directed.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class Debug 
{
	CHECK_NATIVE_NODEOS

#ifdef DEBUG
	/*
	 * NATIVE_DEBUG should provide the following methods:
	 *   print(String)
	 *   println(String)
	 *   printInt(int)
	 *   printIntHex(int)
	 *   printCharArray(char[])
	 *   printStackTrace(Throwable t)
	 *   checkLocked(Object)
	 *   exit()
	 *   beginDebugOutput()
	 *   endDebugOutput()
	 */

/*
 * XXX THIS IS REALLY GROSS.  PLEASE REPLACE WITH ANYTHING.
 */
#define NATIVE_WRAP(FUNC, FORMALARGS, ACTUALARGS) \
	public static void FUNC FORMALARGS { \
			NATIVE_DEBUG.beginDebugOutput(); \
			NATIVE_DEBUG.FUNC ACTUALARGS;    \
			NATIVE_DEBUG.beginDebugOutput(); \
	}
	

	NATIVE_WRAP(print, (String msg), (msg))
	NATIVE_WRAP(println, (String msg), (msg))
	NATIVE_WRAP(printInt, (int i), (i))
	NATIVE_WRAP(printIntHex, (int i), (i))

	/**
	 * Print a file:line location.  Used for asserts and trace
	 * macros.
	 * <p>
	 * Trims the filename down to just the base file name...
	 *
	 * @param file a file name
	 * @param line a line number
	 * @allocation String.substring() may allocate
	 */
	public static void printLocation(String file, int line)
	{
		int off = file.lastIndexOf('/');
		if (off > 0)
			file = file.substring(off+1);

		NATIVE_DEBUG.print(file);
		NATIVE_DEBUG.print(":");
		NATIVE_DEBUG.printInt(line);
	}

	/**
	 * Print a trace message.
	 * <p>
	 * XXX move this into Trace.Java
	 *
	 * @param file the name of the file the TRACE is in
	 * @param line the line number of the file the TRACE is in
	 * @param msg the trace message to print
	 */
	public static void pTrace(String file, int line, String msg) 
	{
		NATIVE_DEBUG.beginDebugOutput();

		printLocation(file, line);
		NATIVE_DEBUG.print(": ");
		NATIVE_DEBUG.println(msg);

		NATIVE_DEBUG.endDebugOutput();
	}

	/**
	 * Get a string of the given Buffer.
	 */
	public static String dbgPacketSummary(BufferHandle bufh)
	{
		return dbgPacketSummary(bufh, 0, bufh.getLength());
	}

	/**
	 * Get a string of maxDump bytes from the given Buffer.
	 */
	public static String dbgPacketSummary(BufferHandle bufh, int len)
	{
		return dbgPacketSummary(bufh, 0, len);
	}

	/**
	 * Get a string of a subsequence of the bytes in the given packet.
	 */
	public static String dbgPacketSummary(BufferHandle bufh, int offset, int length)
	{
		NATIVE_BUFFERHANDLE b;
		String r;
		
		b = bufh.getNativeBufH();
		r = dbgPacketSummary(b, offset, length);
		return r;
	}

	public static String dbgPacketSummary(NATIVE_BUFFERHANDLE b)
	{
		return dbgPacketSummary(b, 0, b.getLength());
	}

	public static String dbgPacketSummary(NATIVE_BUFFERHANDLE b, int len)
	{
		return dbgPacketSummary(b, 0, len);
	}

	public static String dbgPacketSummary(NATIVE_BUFFERHANDLE b, int offset, int len)
	{
		char[] dbgbuf = new char[50 + (len*4)];
		b.dbgPacketSummary(dbgbuf, offset, len);
		return new String(dbgbuf);
	}

	/**
	 * Print the stack trace to the current point of execution in the current
	 * thread.
	 */
	public static void printStackTrace()
	{
		Throwable t = new Throwable();
		NATIVE_DEBUG.printStackTrace(t);
	}

	public static InternalError fatalException(Throwable t, String msg)
	{
		NATIVE_DEBUG.beginDebugOutput();
		NATIVE_DEBUG.print("FATAL EXCEPTION: " +msg);
		NATIVE_DEBUG.printStackTrace(t);
		NATIVE_DEBUG.endDebugOutput();
		NATIVE_DEBUG.exit();
		return new InternalError("exit() returned in Debug.fatalException");
	}

	/**
	 * A "not-implemented print-and-die" function.
	 * Prints the provided string and a backtrace before killing
	 * the VM.
	 *
	 * @param expl the explaination for this rude behavior.
	 */
	public static InternalError notImplemented(String expl)
	{
		NATIVE_DEBUG.beginDebugOutput();
		NATIVE_DEBUG.print("NOT IMPLEMENTED: ");
		NATIVE_DEBUG.println(expl);
		printStackTrace();
		NATIVE_DEBUG.endDebugOutput();
		NATIVE_DEBUG.exit();
		return new InternalError("exit() returned in Debug.notImplemented()");
	}

	/**
	 * A generic panic() function.  Prints the provided string,
	 * a backtrace, and then kills the VM.
	 * <p>
	 * Whee!  This returns an InternalError so that the caller
	 * might throw the result giving the compiler the impression
	 * that dPanic() will never return.  Hehehe.
	 *
	 * @param str The panic message to use as an excuse for killing the machine.
	 */
	public static InternalError doPanic(String str)
	{
		NATIVE_DEBUG.beginDebugOutput();
		NATIVE_DEBUG.print("PANIC: ");
		NATIVE_DEBUG.println(str);
		printStackTrace();
		NATIVE_DEBUG.endDebugOutput();
		NATIVE_DEBUG.exit();
		return new InternalError("exit() returned in Debug.doPanic");
	}
#endif
}
