/*
 * Flow.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import JVM_SYS_PKG.ConfigFile;

import JVM_RESOURCES_PKG.ClassPathResource;

import java.io.IOException;

/**
 * Java wrapper for a flow.  This is the current flow's local
 * interface to its own flow object.
 * <p>
 * If you know BeOS or V, this is like a "Team".  If you prefer Mach
 * (or know what Fluke is), this is similar to a "Task".  If all you
 * know is Unix, then think of this as a "process".  You might also
 * call it a "resource container" or possibly a "protection domain."
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class Flow
	extends NATIVE_FLOW
{
	CHECK_NATIVE_NODEOS

	/**
	 * The Flow ClassLoader for this flow.
	 * <p>
	 * The loader also acts as a flag for a destroyed flow.  If
	 * loader is null, the flow is destroyed.  Note that loader is
	 * null until the first thread in the new flow is started, so
	 * don't rely on destroyed() until the flow actually running.
	 *
	 * @see FlowInitializer#run()
	 */
	private FlowClassLoader loader;

	/**
	 * Cached reference to our kernel held, FlowHandle.
	 */
	PACKAGE FlowHandle kernelHandle;
	
	/**
	 * Create a new Flow, which is a child of the current Flow.
	 * Starts a new thread executing within the new flow.
	 * <p>
	 * This is the only interface for creating a Flow.
	 * <p>
	 * The termHandler will be invoked by a thread in the parent
	 * Flow's if the child Flow is terminated.  This handler has a
	 * short limit for execution.
	 * <p>
	 * If termHandler is null, then no termination handler is called.
	 * <p>
	 * XXX Should be called "create()" to sync back with NodeOS spec.
	 * <p>
	 * XXX don't overload Runnable for the termHandler.  Use a different,
	 * jnodeos interface that doesn't overlap with the other callback interfaces.
	 *
	 * @param cred The principal associated with this Flow
	 * @param tPoolSpec The specification for the new Flow's ThreadPool
	 * @param memSpec The memory limit for this flow
 	 * @param flowStartSpec Class loading and startup information
	 * @param termHandler Termination handler invoked by the NodeOS when
	 * an immediate child flow is being terminated.
	 * @return a FlowHandle onto the new child flow.
	 */
	public FlowHandle newFlow(Credentials cred,
				  ThreadPoolSpec tPoolSpec,
				  MemSpec memSpec,
				  FlowStartSpec flowStartSpec,
				  Runnable termHandler)
	{
		require(cred);
		require(tPoolSpec);
		require(memSpec);
		require(flowStartSpec);
		// XXX require(termHandler);
		
		TRACE(Trace.NODEOS_FLOWSTART, "Flow.newFlow()");

		try
		{
			return super.createFlow(cred,
						tPoolSpec,
						memSpec,
						flowStartSpec,
						termHandler);
		}
		catch (FlowStartException fse)
		{
			// XXX change newFlow signature to throw this.
			Node.println("fse during flow create");
			Node.printStackTrace(fse);

			return null;
		}

		// NOT REACHED
	}

	/**
	 * Just like 'newFlow', but for the root flow which may require
	 * special magic in the backend.
	 *
	 * @see #newFlow
	 */
	PACKAGE static FlowHandle createRootFlow(Credentials cred,
						 ThreadPoolSpec tPoolSpec,
						 MemSpec memSpec,
						 FlowStartSpec flowStartSpec)
		throws FlowStartException
	{
		return NATIVE_FLOW.createRootFlow(cred,
						  tPoolSpec,
						  memSpec,
						  flowStartSpec);
	}

	/**
	 * Initalize a new Flow.  Does not "start" the flow.
	 * Registers the flow in the global FlowRegistry.
	 * <p>
	 * Only invoked via FlowHandle's constructor.  See newFlow
	 * <p>
	 * This is invoked in the context of the new flow by a visiting
	 * thread from the parent flow.
	 *
	 * @param cred the credentials for the new flow. XXX.
	 * @param tPoolspec the thread pool spec for the first thread pool in the new flow
	 * @param memSpec memory spec for the new flow
	 * @param termHandler callback (in parent!) to be invoked when flow is nuked.
	 * @param fcs The set of classes this flow will be allowed to access.
	 * @see #newFlow()
	 * @see FlowHandle
	 */
	PACKAGE Flow(final Credentials cred,
		     final ThreadPoolSpec tPoolSpec,
		     final MemSpec memSpec,
		     final Runnable termHandler,
		     final FlowClassSpec fcs)
	{
		super(cred.getNativeCreds(),
		      tPoolSpec.getNativeSpec(),
		      memSpec.getNativeSpec(),
		      termHandler);

		TRACE(Trace.NODEOS_FLOWSTART, "Flow.<init>(cred=" +cred+ ")");

		this.loader = new FlowClassLoader(this, fcs);

		// register() allocates a new FlowHandle, too
		this.kernelHandle = FlowRegister.register(this);
	}

	/**
	 * Start flow shutdown/cleanup for this flow.
	 * Unregisters flow from the global FlowRegistry.
	 */
	public synchronized void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);

		// FlowRegister.unregister() is synchronized internally
		FlowRegister.unregister(this);

		// loader is the flag for redundant calls to destroy
		if (loader != null)
		{
			loader.destroy();
			loader = null;

			super.destroy();
		}
	}

	public synchronized boolean destroyed()
	{
		// destroyed() is only reliable for threads inside the flow.
		assert(this == Flow.currentFlow());
		
		// XXX ???
		return (loader == null);
	}
	
	/**
	 * Create an explicit thread.  This is a new thread which is
	 * not subject to whatever concurrency limit this Flow's
	 * ThreadPool has.  But, it will be scheduled by the
	 * ThreadPool's scheduler.
	 * 
	 * @allocations: This function will allocate a new Thread object.
	 * @param r the Runnable to start the target thread on.
	 * @param name the String name of the new thread.
	 */
	public void threadStart(java.lang.Runnable r, String name)
	{
		super.threadStart(r, name);
	}

	public Object getObject()
	{
		return super.getObject();
	}

	public void setObject(Object obj)
	{
		super.setObject(obj);
	}

	public String toString()
	{
		return "Flow["
			+ "name=" + this.getName() + ";"
			+ super.toString()
	                + "]";
	}

	/**
	 * Add a new class to an existing flow.
	 */
	public void addClass(String name, byte[] bytecode)
	{
		assert(this == Flow.currentFlow());
		assert(name != null);
		assert(bytecode != null);
		assert(loader);
		loader.addClass(name, bytecode);
	}
	
	/**
	 * Add a package to the list of "system" packages.  System
	 * packages will be loaded by the system class loader.
	 * All other classes will be loaded by the per-flow
	 * classloader.
	 * <p>
	 * All sub-flows *share* the same definition of "system"
	 * packages for now.  This is because only one EE System
	 * is loaded on the NodeOS.  This will have to change.
	 * <p>
	 * Implicitly allows the package.
	 */
	public void addSystemPackage(String name)
	{
		assert(this == Flow.currentFlow());
		assert(name != null);
		assert(loader != null);

		loader.addSystemPackage(name);
	}

	/**
	 * Seal a package so that no other classes can be added to it.
	 *
	 * @param name The name of the package to seal.
	 */
	public void sealPackage(String name)
	{
		assert(this == Flow.currentFlow());
		assert(name != null);
		assert(loader != null);

		loader.sealPackage(name);
	}

	/**
	 * Export a class from the class loader to the flow.
	 *
	 * @param name The name of the class to export.
	 */
	public void exportClass(String name)
	{
		assert(this == Flow.currentFlow());
		assert(name != null);
		assert(loader != null);

		loader.exportClass(name);
	}

	/**
	 * Enable access controls on the class loader.
	 */
	public void enableAccessControls(boolean doit)
	{
		assert(this == Flow.currentFlow());
		assert(loader != null);

		loader.enableAccessControls(doit);
	}

	public Class loadClass(String name)
		throws java.lang.ClassNotFoundException
	{
		require(name != null);
		require(loader);
		
		return this.loader.loadClass(name);
	}

	PACKAGE Class loadClass(String name, boolean resolve)
		throws java.lang.ClassNotFoundException
	{
		require(name != null);
		require(loader);
		
		return this.loader.loadClass(name, resolve);
	}

	/**
	 * XXX Is this evil?  Should it just be loadFile from class path?
	 *
	 * XXX should never return null.
	 */
	public static byte[] readByteCode(String classname)
		throws ClassNotFoundException
	{
		byte[] bytes = null;

		require(classname);

		String cname = classname.replace('.', '/').concat(".class");

		try
		{
			bytes = ClassPathResource.loadFile(cname);
		}
		catch(IOException ioe)
		{
		}

		if (bytes == null)
			throw new ClassNotFoundException("Could not find bytecode for "
							 + classname
							 + ".");
		return bytes;
	}

	public void getStats(FlowStats fs)
	{
		super.getStats(fs.getNative());
	}
	
	/**
	 * Lookup something in the JanosVM config file.
	 * XXX This is a hack to make it easy to get config parameters.
	 */
	public static String lookupConfig(String section,
					  String tag,
					  int instance)
	{
		return ConfigFile.lookup(section, tag, instance);
	}

	/**
	 * Lookup something in the JanosVM config file.
	 * XXX
	 */
	public static String lookupConfig(String section, String tag)
	{
		return lookupConfig(section, tag, 0);
	}
	
	/**
	 * Get the current flow.  The current Flow is the
	 * flow associated with the current Thread.
	 */
	public final static Flow currentFlow()
	{
		NATIVE_FLOW nflow;

		nflow = NATIVE_FLOW.native_currentFlow();
		assert((nflow == null) || (nflow instanceof Flow));
		return (Flow)nflow;
	}

	PACKAGE NATIVE_FLOW getNativeFlow()
	{
		return this;
	}

#ifdef DEBUG
	// always non-null.
	private String name = "Unnamed Flow";
#endif /*DEBUG*/
	
	public void setName(String name)
	{
#ifdef DEBUG
		this.name = name;
#endif /*DEBUG*/	
	}

	public String getName()
	{
#ifdef DEBUG
		return name;
#else
		return null;
#endif /*DEBUG*/
	}
}

