/*
 * FlowClassSpec.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * A specification for which classes a flow can access.
 */
public final class FlowClassSpec
{
	/**
	 * Default set of java classes that should be exported to untrusted
	 * code.
	 */
	public static final String defaultJavaClasses[] = {
		/* Most exceptions... */
		"java.lang.Throwable",
		"java.lang.Exception",
		"java.lang.Error",
		"java.lang.ClassNotFoundException",
		"java.lang.InternalError",
		"java.lang.InterruptedException",
		"java.lang.SecurityException",
		"java.lang.NumberFormatException",
		"java.lang.StringIndexOutOfBoundsException",
		"java.lang.CloneNotSupportedException",

		/* Utility classes */
		"java.lang.Byte",
		"java.lang.Boolean",
		"java.lang.Math",
		"java.lang.Integer",
		"java.lang.Long",
		"java.lang.Object",
		"java.lang.String",
		"java.lang.StringBuffer",
		"java.lang.Cloneable",
		"java.lang.Runnable",

		/* Bah, this one shouldn't be visible... */
		"java.io.IOException",

		"java.util.Hashtable",
		"java.util.Enumeration",
		"java.util.Vector",
	};

	/**
	 * Default set of classes that are considered "system" code.
	 */
	public static final String defaultSystemPackages[] = {
		"java.lang",
		"java.io",
		"java.util",
		JANOS_NODEOS_PKG_S,
		JANOS_UTIL_PKG_S,
		JANOS_ROUTE_PKG_S,
		JANOS_IO_PKG_S,
		JANOS_SERVICES_PKG_S + ".arp",
	};

	/**
	 * Default set of packages that should be "sealed" so that untrusted
	 * code can't add any classes to them.
	 */
	public static final String defaultSealedPackages[] = {
		"java.lang",
		"java.io",
		"java.util",
		JANOS_NODEOS_PKG_S,
		JANOS_ROUTE_PKG_S,
		JANOS_IO_PKG_S,
	};

	/**
	 * Reference to the default spec that should be used by new flows.
	 */
	private static FlowClassSpec defaultSpec;

	/**
	 * The set of packages that contain "system" code.
	 */
	private final Hashtable systemPackages = new Hashtable();

	/**
	 * The set of classes that are safe to export to untrusted code.
	 */
	private final Hashtable exportedClasses = new Hashtable();

	/**
	 * The set of packages which should be "sealed" so that untrusted
	 * code can't add any classes to them.
	 */
	private final Hashtable sealedPackages = new Hashtable();

	public FlowClassSpec()
	{
	}

	public synchronized void addSystemPackage(String pkgName)
	{
		if( !this.isSystemPackage(pkgName) )
		{
			this.systemPackages.put(pkgName, pkgName);
		}
	}

	public synchronized boolean isSystemPackage(String pkgName)
	{
		boolean retval = false;
		
		if( this.systemPackages.get(pkgName) != null )
			retval = true;
		return retval;
	}

	public synchronized void exportClass(String className)
	{
		if( !this.isExportedClass(className) )
		{
			this.exportedClasses.put(className, className);
		}
	}

	public synchronized boolean isExportedClass(String className)
	{
		boolean retval = false;

		if( this.exportedClasses.get(className) != null )
			retval = true;
		return retval;
	}

	public synchronized void sealPackage(String pkgName)
	{
		if( !this.isSealedPackage(pkgName) )
		{
			this.sealedPackages.put(pkgName, pkgName);
		}
	}

	public synchronized boolean isSealedPackage(String pkgName)
	{
		boolean retval = false;
		
		if( this.sealedPackages.get(pkgName) != null )
			retval = true;
		return retval;
	}

	public synchronized Object clone()
	{
		FlowClassSpec fcs = new FlowClassSpec();
		Object retval = fcs;
		Enumeration elems;

		elems = this.systemPackages.elements();
		while( elems.hasMoreElements() )
		{
			fcs.addSystemPackage((String)elems.nextElement());
		}
		elems = this.exportedClasses.elements();
		while( elems.hasMoreElements() )
		{
			fcs.exportClass((String)elems.nextElement());
		}
		elems = this.sealedPackages.elements();
		while( elems.hasMoreElements() )
		{
			fcs.sealPackage((String)elems.nextElement());
		}
		return retval;
	}
		
	PACKAGE static void setDefault(FlowClassSpec fcs)
	{
		/*
		 * XXX Need to get around lazy allocation to stop cross
		 * process pointers.
		 */
		fcs.systemPackages.elements();
		fcs.exportedClasses.elements();
		fcs.sealedPackages.elements();
		FlowClassSpec.defaultSpec = fcs;
	}
	
	PACKAGE static FlowClassSpec getDefault()
	{
		return FlowClassSpec.defaultSpec;
	}
}
