/*
 * FlowInitializer.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.lang.reflect.Constructor;

import JVM_SYS_PKG.LocalTeam;

import JVM_RESOURCES_PKG.GCPolicy;

/**
 * Callback for initializing a new Flow (in the context of one of its
 * own threads).
 * <p>
 * This class's <code>run()</code> should be a method on a Flow
 * object, but starting a thread requires a <emph>public</emph>
 * <code>run()</code> method, which would be visible to the world on
 * Flow.
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class FlowInitializer
	implements Runnable
{
	/**
	 * The spec that determines how the flow should be started up.
	 */
	private final FlowStartSpec fss;
	
	/**
	 * Must be called in the context of the flow for which
	 * it is being used.
	 */
	PACKAGE FlowInitializer(FlowStartSpec fss)
	{
		/* Make a copy in this flow's memory */
		this.fss = (FlowStartSpec)fss.clone();
	}

	/**
	 * The run() function that the flow initializing thread will be
	 * executing.
	 */
	public void run()
	{
		Flow flow;

		TRACE(Trace.NODEOS_FLOWSTART, "FlowInitializer.run()");

		flow = Flow.currentFlow();
		assert(flow != null);

		try
		{
			GCPolicy gp;

			/* Start up the GC threads within this Flows context */
			gp = GCPolicy.currentPolicy();
			if( (gp != null) && !gp.startThreads() )
			{
				Node.println("Unable to start up GC threads in new flow.");
				flow.destroy();
			}
			/* Start the startclass in the flow. */
			startInitialClass(flow, this.fss);
		}
		catch (Throwable e) // XXX
		{
			Node.println("FlowInitialzier.run(): Trying to start "
				     + this.fss.getStartClass()
				     + " in new flow(" +flow+ ") and got:");
			Node.printStackTrace(e);

			flow.destroy();
		}
	}
	
	/**
	 * Start the flow on its way.
	 */
	private static void startInitialClass(Flow flow,
					      FlowStartSpec fss)
		throws Exception
	{
		FlowRun flowRun;
		FlowClassLoader loader;

		require(flow);
		require(fss.getStartClass());
		require(LocalTeam.fetch().owns(fss.getStartClass()));
		require(fss.getArguments());
		
		TRACE(Trace.NODEOS_FLOWSTART, "FlowInitializer.startInitialClass(" +fss.getStartClass()+ ")");

		Class c;
		
		c = flow.loadClass(fss.getStartClass(), true);
		flowRun = createStartClassInstance(c);
		
		TRACE(Trace.NODEOS_FLOWSTART, "FlowInitializer.startInitialClass(): running startup class.");
		flowRun.run(flow,
			    fss.getArguments(),
			    fss.getCommSpaceHandle());
	}
	
	private static FlowRun createStartClassInstance(Class c)
		throws Error,  // XXX Error should be a Janos-specific Error
		NoSuchMethodException,
		IllegalAccessException,
		InstantiationException,
		java.lang.reflect.InvocationTargetException // XXX XXX XXX
	{
		Object instance;

		require(c);

		/*
		 * This seemed so easy:
		 *    flow.flowRun = (FlowRun) c.newInstance();
		 *
		 * But no, it's not, because we don't want to require
		 * c to have a public constructor.  See
		 * ClassInstantiationKey.
		 *
		 * Or, we need to have a VM method that will execute a
		 * private/package constructor on the class.  We trust
		 * the EE runtime to only pass safe classes.
		 */
		try
		{
			Class[] paramTypes;
			Object[] params;
			Constructor con;

			paramTypes = new Class[]{ ClassInstantiationKey.class };
			con = c.getConstructor(paramTypes);
							   
			params = new Object[] { new ClassInstantiationKey(c) };

			instance = con.newInstance(params);
		}
		catch(NoSuchMethodException e)
		{
			// Try the no-arg public constructor version(?)
			instance = c.newInstance();
		}

		if (! (instance instanceof FlowRun))
			throw new Error(c.getName() + " is not an instanceof FlowRun.");

		return (FlowRun)instance;
	}
}
