/*
 * FlowRegister.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX

import JVM_SYS_PKG.LocalTeam;
import JVM_SYS_PKG.Visitor;
import JVM_SYS_PKG.CarrierVisitor;
import JVM_SYS_PKG.DeadTeamException;

/**
 * Maintain a map of all the Flows in the system.  Currently
 * only used by the FlowInfo InfoHook.
 * <p>
 * <b>NOTE:</b> We don't use a hashtable to map flows
 * from name to flow object, becuase names are mutable and
 * are set after the flow is registered.
 * <p>
 * XXX currently this class is only useful for debugging builds.
 * <p>
 * XXX rename to FlowRegistry
 *
 * @see Flow#newFlow()
 * @see Flow#destroy()
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class FlowRegister
{
	/**
	 * This is a real global variable.  It is maintained
	 * in the kernel.
	 */
	private static final Vector flows = new Vector(30);

	/**
	 * InfoHook class that reports the current set of flows in the
	 * system via the NodeOS Status interface.
	 */
	private static class FlowInfo
		extends InfoHook
	{
		private FlowInfo()
		{
			super("Flow");
		}

		public String[] getInfo()
		{
			String retval[];

			synchronized(flows)
			{
				retval = new String[flows.size()];
				
				for (int i = 0; i < flows.size(); i++)
				{
					FlowHandle elem =
						(FlowHandle)flows.elementAt(i);
					retval[i] = elem.toString();
				}
			}
			return retval;
		}
	}

	static {
		Status.addInfoHook(new FlowInfo());
	}

	/**
	 * Register the given flow.
	 *
	 * @allocation: a new FlowHandle to hold a reference to the flow
	 * @allocation: a damned Visitor
	 */
	PACKAGE static FlowHandle register(final Flow flow)
	{
		FlowHandle retval = null;
		
		// Switch into kernel context, create a FlowHandle pointing
		// to the given flow and register the handle in the table.

		try
		{
			CarrierVisitor cv;
			
			LocalTeam.fetch().getKernelTeam().visit(cv = new CarrierVisitor() {
				public void run() throws Exception {
					FlowHandle fh = new FlowHandle(flow);
					synchronized(flows)
					{
						flows.addElement(fh);
					}
					this.value = fh;
				}
			});
			retval = (FlowHandle)cv.value;
		}
		catch (DeadTeamException dte)
		{
		}
		catch (Exception whatever)
		{
			FATALEXCEPTION(whatever, "Exception in FlowRegister.register()");
		}
		return retval;
	}

	/**
	 * Unregister a flow.
	 * <p>
	 * NOTE: this method is invoked in a context where most of the
	 * Flow has been destroyed.  Beware.
	 * <p>
	 * XXX this is bogus-slow.  Could cache index in flow object?
	 */
	PACKAGE static void unregister(final Flow flow)
	{
		TRACE(Trace.NODEOS_DESTROY, "Unregistering flow " +flow);

		try
		{
			LocalTeam.fetch().getKernelTeam().visit(new Visitor() {
				public void run() throws Exception {
					synchronized(flows)
					{
						flows.removeElement(flow.kernelHandle);
					}
				}
			});
		}
		catch (DeadTeamException dte)
		{
			/* XXX Bah, we'll leak the FlowHandle... */
		}
		catch (Exception whatever)
		{
			FATALEXCEPTION(whatever, "Exception in FlowRegister.unregister()");
		}
	}

#ifdef DEBUG
	/**
	 * Return the flow object corresponding to the given name
	 */
	public static FlowHandle findFlow(String name)
	{
		FlowHandle retval = null;
		int i;

		for( i = 0; (i < flows.size()) && (retval == null); i++ )
		{
			FlowHandle elem = (FlowHandle)flows.elementAt(i);
			if( elem.getName().equals(name) )
				retval = elem;
		}
		return retval;
	}
#endif
}
