/*
 * InChannel.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;

/**
 * Java wrapper for a NodeOS InChannel.
 *
 * @see CutThroughChannel
 * @see OutChannel
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class InChannel
	extends NATIVE_INCHANNEL
{
	CHECK_NATIVE_NODEOS

	/**
	 * Create a new InChannel.
	 */
	public static InChannel create(Flow flow,
				       ThreadPoolSpec tps,
				       DemultiplexKey classifier,
				       String protocolSpec,
				       String addrSpec,
				       DeliverFunc deliverFunc)
		throws IOException,
		       PacketAddressFormatException,
		       ProtocolSpecFormatException
	{
		// XXX doesn't sufficiently protect the kernel...
		if((flow == null)
		   || (tps == null)
		   || (classifier == null)
		   || (protocolSpec == null)
		   || (deliverFunc == null))
			throw new NullPointerException();

		TRACE(Trace.NODEOS_INCOMING, "new InChannel(cl="+classifier+"; protocol='"+protocolSpec+"'; addr='"+addrSpec+"')");

		return new InChannel(flow,
				     tps,
				     classifier,
				     protocolSpec,
				     addrSpec,
				     deliverFunc);
	}

	/**
	 * XXX flow parameter is silly.
	 */
	public static InChannel create(Flow flow,
				       DemultiplexKey classifier,
				       String protocolSpec,
				       String addrSpec,
				       DeliverFunc deliverFunc)
		throws IOException,
		       PacketAddressFormatException,
		       ProtocolSpecFormatException
	{
		return InChannel.create(flow,
					ThreadPoolSpec.getDefault(),
					classifier,
					protocolSpec,
					addrSpec,
					deliverFunc);
	}
	

	private InChannel(Flow flow,
			  ThreadPoolSpec tps,
			  DemultiplexKey classifier,
			  String protocolSpec,
			  String addrSpec,
			  DeliverFunc deliverFunc)
		throws IOException,
		       PacketAddressFormatException,
		       ProtocolSpecFormatException
	{
		super(flow.getNativeFlow(),
		      tps.getNativeSpec(),
		      classifier.getNativeKey(),
		      protocolSpec,
		      addrSpec,
		      deliverFunc);
	}
	
	/**
	 * Add a buffer to the InChannel
	 */
	public void addBuffer(BufferHandle bufh)
	{
		NATIVE_BUFFERHANDLE nb;

		if (bufh == null)
			throw new NullPointerException();

		bufh.systemOwns();
		
		nb = bufh.getNativeBufH();
		super.native_addBuffer(nb);
	}

	/**
	 * Remove a buffer from the queue of free buffers
	 * on this InChannel.  Returns null if there are
	 * no free buffers on this InChannel.
	 */
	public BufferHandle remBuffer()
	{
		return (BufferHandle)super.native_remBuffer();
	}
	
	/**
	 * Add `count' buffers to the in channel.
	 *
	 * @allocation Creates 'count' new BufferHandles (and buffers).
	 */
	public void addBuffers(int count)
	{
		int lpc;
		
		require(count > 0);

		for( lpc = 0; lpc < count; lpc++ )
		{
			BufferHandle bh;
			bh = BufferHandle.create();
			bh.systemOwns();
			super.native_addBuffer(bh.getNativeBufH());
		}
	}

	/**
	 * Remove `count' buffers from the channel.  Or all the buffers
	 * if there are less than `count'.
	 */
	public void remBuffers(int count)
	{
		int lpc;
		
		require(count > 0);

		for( lpc = 0; lpc < count; lpc++ )
		{
			Object o;
			o = super.native_remBuffer();
			if (o == null)
				return;
			require(o instanceof BufferHandle);
		}
	}

	/**
	 * @param cs The object to place the channel stats in.
	 */
	public void getStats(ChanStats cs)
	{
		super.getStats(cs.getNative());
	}
	
	public String toString()
	{
		return super.toString();
	}

	/**
	 * Package access function for directly feeding a packet to an
	 * in channel.
	 *
	 * @param bh The buffer handle holding the packet.
	 * @return True if the packet was properly dispatched, false otherwise.
	 */
	PACKAGE boolean handoff(BufferHandle bh)
	{
		assert(bh != null); // XXX return true?  throw error?
		return super.handoff(bh.getNativeBufH());
	}

	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		super.destroy();
	}
	
	PACKAGE NATIVE_INCHANNEL getNativeInChan()
	{
		return this;
	}

	protected void finalize()
	{
		destroy(); // XXX ???
	}
}
