/*
 * Main.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

IMPORT_PROFILER

/**
 * Boot the Janos Java NodeOS.  Initializes the underlying native
 * NodeOS, sets up the basic system packages in the ClassLoader, and
 * creates the root flow.
 * <p>
 * The Janos Java NodeOS must always startup with this class.  It doesn't
 * do much besides initialize the NodeOS and then fire up its first
 * argument in the root flow.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class Main
{
	/**
	 * All Janos active routers start with this function.  We load
	 * up native libraries, define the set of "system" packages,
	 * and fire up the root flow.  The arguments to main() should
	 * be a class to instantiate in the root flow and whatever
	 * arguments it wants.
	 *
	 * @see FlowClassLoader for info on system packages.
	 */
	public static void main(String[] args)
	{
		NATIVE_SETUP.initialize();

		PROFILESTAGE("SysInit");

		Node.println("Janos Java NodeOS v1.1.5");
		
		require(args);
		if (args.length == 0)
		{
			Node.println("ERROR: Must supply a class to load into root flow.");
			Node.shutdown();
		}

		String initClass;
		String[] initArgs;
		String initPackage;
		int pkgend;
		
		initClass = args[0];
		
		// Find the package the initial class is in.
		pkgend = initClass.lastIndexOf('.');
		initPackage = null;
		if (pkgend != -1)
			initPackage = initClass.substring(0, pkgend);
		defineSystemPackages(initPackage);

		initArgs = new String[args.length - 1];
		if (args.length > 1)
			System.arraycopy(args, 1, initArgs, 0, initArgs.length);
		
		PROFILESTAGE("NodeOSInit");

		try
		{
			startRootFlow(initClass, initArgs);
		}
		catch (FlowStartException fse)
		{
			Node.println("Exception during root flow setup: ");
			Node.println("Cause: ");
			Node.printStackTrace(fse.getCause());
			Node.println("FlowStartException:");
			Node.printStackTrace(fse);
			Node.shutdown();
		}

		/* XXX don't do anything if startRootFlow returns... */
	}

	/**
	 * Start the native root flow.  Not really the "root flow" in
	 * Moab's case, as we start the root flow in md.c before Kaffe's
	 * main even gets going.
	 */
	private static void startRootFlow(String initClass,
					  String[] initArgs)
		throws FlowStartException
	{
		require(initClass);
		require(initArgs);

		Flow.createRootFlow(new Credentials("root cred"),
				    new ThreadPoolSpec(3, 32*1024,
						       ThreadScheduler.
						       basicScheduler()),
				    new MemSpec(),
				    new FlowStartSpec(initClass, initArgs, null));
	}

	/**
	 * Define the set of "system" packages that are loaded into
	 * the system flow.  This set is basically fixed, except for
	 * the implicit package passed on the command line.
	 */
	private static void defineSystemPackages(String initPackage)
	{
		FlowClassSpec fcs = new FlowClassSpec();
		int lpc;

		/* Set up the defaults */
		for( lpc = 0;
		     lpc < FlowClassSpec.defaultSystemPackages.length;
		     lpc++ )
		{
			fcs.addSystemPackage(FlowClassSpec.
					     defaultSystemPackages[lpc]);
		}
		for( lpc = 0;
		     lpc < FlowClassSpec.defaultJavaClasses.length;
		     lpc++ )
		{
			fcs.exportClass(FlowClassSpec.defaultJavaClasses[lpc]);
		}
		for( lpc = 0;
		     lpc < FlowClassSpec.defaultSealedPackages.length;
		     lpc++ )
		{
			fcs.sealPackage(FlowClassSpec.
					defaultSealedPackages[lpc]);
		}
		/* Add the initial package */
		if (initPackage != null)
			fcs.addSystemPackage(initPackage);
		FlowClassSpec.setDefault(fcs);
	}
}

