/*
 * OutChannel.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.io.IOException;

import java.net.DatagramSocket;
import java.net.DatagramPacket;

/**
 * Java wrapper for a NodeOS OutChannel.
 *
 * @see CutThroughChannel
 * @see InChannel
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class OutChannel
	extends NATIVE_OUTCHANNEL
{
	CHECK_NATIVE_NODEOS

	/**
	 * Create a new OutChannel.
	 */
	public static OutChannel create(Flow flow,
					String protocolSpec,
					String addressSpec)
		throws IOException,
		PacketAddressFormatException,
		ProtocolSpecFormatException
	{
		if ((flow == null)
		    || (protocolSpec == null)
		    || (addressSpec == null))
			throw new NullPointerException(/*XXX*/);

		return new OutChannel(flow,
				      new NetworkSpec(),
				      protocolSpec,
				      addressSpec);
	}

	public static OutChannel create(Flow flow,
					NetworkSpec ns,
					String protocolSpec,
					String addressSpec)
		throws IOException,
		PacketAddressFormatException,
		ProtocolSpecFormatException
	{
		if ((flow == null)
		    || (ns == null)
		    || (protocolSpec == null)
		    || (addressSpec == null))
			throw new NullPointerException(/*XXX*/);

		return new OutChannel(flow, ns, protocolSpec, addressSpec);
	}
	
	private OutChannel(Flow flow,
			   NetworkSpec ns,
			   String protocolSpec,
			   String addressSpec)
		throws IOException,
		       PacketAddressFormatException,
		       ProtocolSpecFormatException
	{
		super(flow.getNativeFlow(), ns, protocolSpec, addressSpec);
	}

	/**
	 * Send a single buffer.  This is a synchronous send.  The buffer
	 * will be owned by the system for the duration of the send,
	 * and then ownership will be restored to the caller after the send.
	 */
	public boolean send(BufferHandle bh)
	{
		NATIVE_BUFFERHANDLE nbh;
		boolean retval;
		
		bh.systemOwns();
		nbh = bh.getNativeBufH();
		retval = super.send(nbh);
		bh.userOwns();
		return retval;
	}

	/**
	 * Send a single buffer.  Recycle the buffer (by making it
	 * available to the given InChannel after the send is complete.
	 * This is a synchronous send.  (It should probably be
	 * async.)
	 */
	public void sendAndRecycle(BufferHandle bh, InChannel dustbin)
	{
		NATIVE_BUFFERHANDLE nbh;

		bh.systemOwns();
		nbh = bh.getNativeBufH();
		super.sendAndRecycle(nbh, dustbin.getNativeInChan());
		bh.userOwns();
	}

	public void sendAsync(BufferHandle bh, DeliverFunc callback)
	{
		NOTIMPLEMENTED("OutChannel.sendAsync()");
	}
	
	/**
	 * Send an array of buffer handles.
	 *
	 * @allocation Allocates a new PacketBuffer and a new byte array
	 * and linearizes into that buffer.
	 */
	public void sendv(BufferHandle[] bh)
	{
		NOTIMPLEMENTED("OutChannel.sendv()");
	}

	/**
	 * @param cs The object to place the channel stats in.
	 */
	public void getStats(ChanStats cs)
	{
		super.getStats(cs.getNative());
	}
	
	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		super.destroy();
	}

	public String toString()
	{
		return super.toString();
	}
}
