/*
 * ProtocolSpec.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/**
 * NodeOS protocol spec parser, it takes formatted strings and splits apart
 * their protocol components.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class ProtocolSpec
{

	/**
	 * A handy function for quickly figuring out what the last component
	 * of a spec is.  This is preferred over creating and parsing the spec
	 * only to get its last component.
	 *
	 * @param spec A protocol spec.
	 * @return The last component of the spec.
	 */
	PACKAGE static String lastComponent(String spec)
		throws ProtocolSpecFormatException
	{
		String retval;
		int idx;

		idx = spec.lastIndexOf('/');
		if( idx == 0 )
			throw new ProtocolSpecFormatException("Can't start with a '/'");
		else if( idx == -1 )
			retval = spec;
		else
			retval = spec.substring(idx + 1);
		return retval;
	}
	
	/**
	 * Contains the strings that make up the protocol format
	 */
	private final Vector components = new Vector();

	/**
	 * Construct an empty spec
	 */
	PACKAGE ProtocolSpec()
	{
	}

	/**
	 * Construct a spec from a formatted string
	 */
	PACKAGE ProtocolSpec(String spec)
		throws ProtocolSpecFormatException // XXX where?
	{
		StringTokenizer st = new StringTokenizer(spec, "/");
		
		try
		{
			String protocol;
			
			while( true )
			{
				protocol = st.nextToken();
				this.components.addElement(protocol);
			}
		}
		catch(NoSuchElementException e)
		{
		}
	}

	/**
	 * Return the number of components in the protocol spec
	 */
	PACKAGE int componentCount()
	{
		return this.components.size();
	}

	/**
	 * Return the protocol string for a specific index
	 */
	PACKAGE String getComponent(int index)
	{
		return (String)this.components.elementAt(index);
	}

	/**
	 * Add a protocol string component
	 */
	PACKAGE void addComponent(String protocol)
	{
		this.components.addElement(protocol);
	}

	public boolean equals(Object o)
	{
		boolean retval = false;
		
		if (o instanceof ProtocolSpec)
		{
			ProtocolSpec ps = (ProtocolSpec)o;
			int thisSize, psSize;

			thisSize = this.components.size();
			psSize = ps.components.size();
			if( thisSize == psSize )
			{
				int lpc;
				
				for( retval = true, lpc = 0;
				     (lpc < thisSize) && retval;
				     lpc++ )
				{
					retval = this.components.
						elementAt(lpc).equals(
							ps.components.
							elementAt(lpc));
				}
			}
		}
		return retval;
	}

	public String toString()
	{
		StringBuffer retval;
		int lpc, size;

		size = this.components.size();
		retval = new StringBuffer(size * 5);
		for( lpc = 0; lpc < size; lpc++ )
		{
			if( lpc > 0 )
				retval.append("/");
			retval.append(this.components.elementAt(lpc).
				      toString());
		}
		return retval.toString();
	}
}
