/*
 * DebugControl.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.StringTokenizer;  // XXX
import java.util.Hashtable;

import java.io.PrintStream; // XXX

/**
 * Control for the trace macros spread through-out creation.
 * <p>
 * TRACE() macros include a Trace flag that determines when the trace
 * macro should print its string.  The set of flags that are enabled
 * is determined by the system property "janos.debugFlags".
 * <p>
 * XXX Use intern() and ==.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class Trace
{
	/**
	 * The set of bit flags describing which trace flags
	 * are on and which are off.
	 */
	private static long flags;

	/**
	 * Default trace flags.  For a "production" environment
	 * this could be null/empty.
	 */
	private static final String defaultFlags = "";

	/**
	 * The property to look at to find flags at run time.
	 */
	private static final String flagPropKey  = "janos.debugFlags";

	/**
	 * The maximum number of Trace flags that are supported.
	 * Could be bigger if BitSet was used.
	 */
	private static final int MAX_FLAGS = 64; // # bits in long

	/**
	 * An array of string names.  The index of the name
	 * is the *bit offset* of the flag in the <code>flags</code>
	 * set.
	 */
	private static final String[] names = new String[MAX_FLAGS];

	/**
	 * A mapping of string names to Long objects.  Each Long
	 * contains the set of bits the name represents.
	 */
	private static final Hashtable compositeFlags = new Hashtable();

	/**
	 * Control debugging of this class.  Obviously, this isn't
	 * set via one of the debug flags.
	 */
	private static boolean debugTrace = false;


// Define a flag.  name[idx] = "NAME"; and NAME = (1L << IDX)
#define FLAG(NAME, IDX)					\
	static						\
	{						\
		names[IDX] = #NAME;			\
	}						\
	public static final long NAME = (1L<<IDX)
// end #define

		
// Define a composite flag
#define COMPOSITEFLAG(NAME, VAL)				\
	static							\
	{							\
		compositeFlags.put(#NAME, new Long(VAL));	\
	}
// end #define
		
	// There are only 64 flags available.

	// Janos NodeOS flags
	FLAG(NODEOS_INCOMING, 1);
	FLAG(NODEOS_BUFH,     2);
	FLAG(NODEOS_OUTGOING, 3);
	FLAG(NODEOS_DISPATCH, 4);
	FLAG(NODEOS_CLASSLOAD,5);
	FLAG(NODEOS_DESTROY  ,6);
	FLAG(NODEOS_THREADS  ,7);
	FLAG(NODEOS_FLOWSTART,8);
	FLAG(NODEOS_COMMSPACE,9);

	// Janos Route flags
	FLAG(ROUTE_UPDATE, 10);

	/**
	 * Initailze the debug flags from the 'janos.debugFlags' system
	 * property.
	 * <p>
	 * Static initializers are executed in textual order, so this goes
	 * after the intialization of the various flag names (above).
	 */
	static 
	{
		String s;
		s = java.lang.System.getProperty(flagPropKey, defaultFlags);

		if (debugTrace)
			System.out.println("TRACE: " +flagPropKey+ " is set to " +s);

		setFlags(s);

		if (debugTrace)
			System.out.println("TRACE: final trace mask is 0x" +Long.toHexString(flags));
	}

	public static String[] getNameList()
	{
		return names;
	}
	
	public static void setFlags(String flagsStr)
	{
		setFlags(System.out, flagsStr);
	}

	public static void setFlags(PrintStream out, String flagsStr)
	{
		StringTokenizer st = new StringTokenizer(flagsStr,",|;");
	TOKENIZING:
		while(st.hasMoreTokens())
		{
			String flag = st.nextToken();

			if (flag.equals("ALL"))
			{
				flags = -1L;
				return;
			}

			if (flag.equals("NONE"))
			{
				flags = 0L;
				return;
			}

			// Check if its a regular debug flag
			for (int i = 0; i < MAX_FLAGS; i++)
			{
				if (flag.equals(names[i]))
				{
					setFlag(i);

					if (debugTrace)
					{
						out.println("TRACE: " +flag+ " is set (index " +i+ ").");
						out.println("TRACE: trace mask is now " +Long.toBinaryString(flags));
					}
					
					continue TOKENIZING;
				} 
			}

			// Check if its a composite debug flag
			Object o;
			o = compositeFlags.get(flag);
			if (o != null)
			{
				long val;
				
				assert(o instanceof Long);
				val = ((Long)o).longValue();
				flags |= val;
				if (debugTrace)
				{
					out.println("TRACE: " +flag+ " is a composite flag.");
					out.println("TRACE: trace mask is now " +Long.toBinaryString(flags));
				}
				continue TOKENIZING;
			}

			out.println("WARNING: Trace flag \"" +flag+ "\" is not valid.");
		}
	}

	public static final boolean check(long mask)
	{
		return ((flags & mask) != 0L);
	}

	private static final void setFlag(int idx)
	{
		flags |= (long)(1L<<idx);
	}

}
