/*
 * moab_Flow.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector;

import JVM_SYS_PKG.Team;
import JVM_SYS_PKG.LocalTeam;
import JVM_SYS_PKG.CarrierVisitor;
import JVM_SYS_PKG.Importable;
import JVM_SYS_PKG.Exportable;
import JVM_SYS_PKG.PersistentExport;
import JVM_SYS_PKG.ThrowInVisitException;
import JVM_SYS_PKG.DeadTeamException;
import JVM_RESOURCES_PKG.GCPolicy;
import JVM_RESOURCES_PKG.StdioSpec;
import JVM_RESOURCES_PKG.MemorySpec;
import JVM_RESOURCES_PKG.ResourceSpec;
import JVM_RESOURCES_PKG.ResourceSpecException;

/**
 * Java wrapper for a NodeOS Flow.
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class moab_Flow
	extends Exportable
	implements PersistentExport
{
	CHECK_NATIVE_NODEOS

	private final Runnable terminationHandler;
	private final moab_ThreadPool threadPool;
	private final Vector inChannels = new Vector();
	private final Vector outChannels = new Vector();

	private Object obj;

	private static int teamId = 0;

	/**
	 * The "root" flow termination handler.  This is invoked
	 * when the "root" flow is terminated, and causes the
	 * NodeOS to initiate shutdown.
	 * <p>
	 * The Java "root" flow is not the actual root flow.  (The
	 * real root flow is C-only and is created in janosvm/config/i386/nodeos/md.c.)
	 */
	private static class RootFlowTermHandler
		implements java.lang.Runnable
	{
		public void run()
		{
			Node.shutdown();
		}
	}

	
	// XXX 
	private static String genTeamName(Credentials cred,
					  FlowStartSpec fss)
	{
		int id;

		synchronized(moab_Flow.class)
		{
			id = teamId++;
		}

		return cred.toString() + "-" + id;
	}

	/**
	 * Create a new flow in a new Team.
	 *
	 * @see Flow#newFlow()
	 */
	PACKAGE static FlowHandle createFlow(final Credentials cred,
					     final ThreadPoolSpec tPoolSpec,
					     final MemSpec memSpec,
					     final FlowStartSpec flowStartSpec,
					     final Runnable termHandler)
		throws FlowStartException
	{
		ResourceSpec[] resources;
		Flow f;
		
		try
		{
			resources = newFlowResources(memSpec);

			String teamName = genTeamName(cred, flowStartSpec);
			Team childTeam = new Team(teamName, resources);
			CarrierVisitor visitFunc = new CarrierVisitor() {
				public void run() throws Exception {
					Flow child;
					
					child = new Flow(cred,
							 tPoolSpec,
							 memSpec,
							 termHandler,
							 FlowClassSpec.getDefault());
					
					child.threadStart(new FlowInitializer(flowStartSpec),
							  "Initializer");
					
					// "return" the child pointer to the calling Team
					this.value = child;
				}
			};

			// Do the actual visit, creating the new flow
			// and starting a thread
			childTeam.visit(visitFunc);
			
			// Store the cross-process child Flow pointer
			// in a new FlowHandle
			return new FlowHandle((Flow)(visitFunc.value));
		}
		catch (ResourceSpecException rse)
		{
			throw new FlowStartException("Resource spec failure",
						     rse);
		}
		catch (DeadTeamException dte)
		{
			throw new FlowStartException("Child team died too soon", dte);
		}
		catch (ThrowInVisitException tive)
		{
			throw new FlowStartException("Exception during bootstrap visit", tive);
		}

		// NOT REACHED
	}

	/**
	 * Create root flow.  Different from createFlow because we
	 * use the current Team to host the root flow.  Also the
	 * root flow has no termination handler (where would it run?)
	 *
	 * @see Flow#newFlow()
	 */
	PACKAGE static FlowHandle createRootFlow(final Credentials cred,
						 final ThreadPoolSpec tPoolSpec,
						 final MemSpec memSpec,
						 final FlowStartSpec flowStartSpec)
		throws FlowStartException
	{
		ResourceSpec[] resources;
		Flow f;
		
		try
		{
			resources = newFlowResources(memSpec);

			Team kTeam = LocalTeam.fetch().getKernelTeam();
			CarrierVisitor visitFunc = new CarrierVisitor() {
				public void run() throws Exception {
					Flow child;
					
					child = new Flow(cred,
							 tPoolSpec,
							 memSpec,
							 new RootFlowTermHandler(),
							 FlowClassSpec.getDefault());
					
					DEBUGONLY(child.setName("Root Flow"));

					// XXX cloning the flowStartSpec isn't necessary
					// for the root flow (same team).
					child.threadStart(new FlowInitializer(flowStartSpec),
							  "Root Flow Initializer");
					
					// "return" the child pointer to the calling Team
					this.value = child;
				}
			};

			// Do the actual visit, creating the new flow
			// and starting a thread
			kTeam.visit(visitFunc);

			// Store the cross-process child Flow pointer
			// in a new FlowHandle
			return new FlowHandle((Flow)(visitFunc.value));
		}
		catch (ResourceSpecException rse)
		{
			FATALEXCEPTION(rse, "Resource exception when creating root flow");
		}
		catch (DeadTeamException dte)
		{
			FATALEXCEPTION(dte, "Kernel team is dead.  I think not!");
		}
		catch (ThrowInVisitException tive)
		{
			FATALEXCEPTION(tive, "Exception while creating team");
		}

		// NOT REACHED
	}


	private static ResourceSpec[] newFlowResources(MemSpec memSpec)
		throws ResourceSpecException
	{
		StdioSpec stdioSpec;
		ResourceSpec[] resources;
		Team kernelTeam = LocalTeam.fetch().getKernelTeam();

		// XXX lameo spec, should be 'none' on all counts?
		// XXX make this static
		stdioSpec = new StdioSpec("builtin://stdin",
					  "builtin://stdout",
					  "builtin://stderr");
		resources = new ResourceSpec[] {
			stdioSpec,
			moab_Setup.getImportedNameSpaces(),
			new MemorySpec(memSpec.getSize(),
				       MemoryPolicy.class.getName())
		};

		return resources;
	}

	/**
	 * Initalize a new Flow.  See newFlow().
	 * <p>
	 * Must be run in the context of the flow's team.
	 */
	PACKAGE moab_Flow(moab_Credentials cred,
			  moab_ThreadPoolSpec threadPoolSpec,
			  moab_MemSpec memSpec,
			  Runnable termHandler)
	{
		// XXX all parameters are pointers into parent flow's heap.  Beware!

		this.terminationHandler = termHandler;
		this.threadPool = new moab_ThreadPool(threadPoolSpec);
		init(cred, this.threadPool, memSpec);
	}

	private native void init(moab_Credentials cred,
				 moab_ThreadPool threadPool,
				 moab_MemSpec memSpec);

	/**
	 * Called from native code (in moab_Flow.c) when a flow's
	 * termination handler is invoked by Moab.
	 * <p>
	 * Note that this will be invoked in the context of the parent
	 * flow.
	 */
	PACKAGE void native_termHandler() 
	{
		if (terminationHandler != null)
			terminationHandler.run();
	}

	PACKAGE synchronized void addChannel(moab_InChannel ic)
	{
		this.inChannels.add(ic);
	}

	PACKAGE synchronized InChannel getInChannel(int index)
	{
		return (InChannel)this.inChannels.elementAt(index);
	}

	PACKAGE synchronized void remChannel(moab_InChannel ic)
	{
		this.inChannels.remove(ic);
	}

	PACKAGE synchronized void addChannel(moab_OutChannel oc)
	{
		this.outChannels.add(oc);
	}

	PACKAGE synchronized OutChannel getOutChannel(int index)
	{
		return (OutChannel)this.outChannels.elementAt(index);
	}

	PACKAGE synchronized void remChannel(moab_OutChannel oc)
	{
		this.outChannels.remove(oc);
	}

	PACKAGE native void getStats(moab_FlowStats fs);
	
	PACKAGE native void threadStart(java.lang.Runnable r, String name);

	PACKAGE native void destroy0();

	PACKAGE void destroy()
	{
		// destroy the Moab flow
		// XXX destroy0();
	}

	PACKAGE Object getObject()
	{
		return this.obj;
	}

	PACKAGE void setObject(Object obj)
	{
		this.obj = obj;
	}

	public String toString()
	{
		return "obj=" +(obj != null ? obj.toString() : "(null)")+ ";"
			+ "inChan-count=" +inChannels.size()+ ";"
			+ "outChan-count=" +outChannels.size();
	}

	PACKAGE static native moab_Flow native_currentFlow();
}
