/*
 * moab_Thread.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

/**
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class moab_Thread
{
	CHECK_NATIVE_NODEOS

	/**
	 * Per-thread state.
	 * <p>
	 * This is part of the NodeOS spec.  Probably doesn't need to
	 * be part of the Java-level interface (users of this should
	 * just subclass Janos Thread).
	 */
	private Object obj;
	
	/**
	 * The ThreadPool this thread is a part of.
	 * <p>
	 * XXX CURRENTLY THREADPOOL IS INTERNAL TO MOAB.
	 */
	// private final moab_ThreadPool threadPool;

	/**
	 * The real underlying thread object.
	 */
	private final ThreadWrapper jlThread;

	/**
	 * Hide java.lang.Thread in a member class so that subclasses
	 * of moab_Thread don't inherit the public java.lang.Thread
	 * methods.
	 */
	private class ThreadWrapper
		extends java.lang.Thread
	{
		private ThreadWrapper(String name)
		{
			super(name == null ? "no-name moab_Thread wrapper": name);
		}

		public void run()
		{
			panic("moab_Thread.ThreadWrapper.run() should never be invoked.");
		}

		/**
		 * Return container class.  Used to convert from
		 * java.lang.Thread.currentThread() to a moab_Thread().
		 */
		private moab_Thread moab_Thread()
		{
			return moab_Thread.this;
		}
	}

	/**
	 * InfoHook class that reports on all of the native threads
	 * in the system.
	 */
	private static class moab_ThreadInfo
		extends InfoHook
	{
		private moab_ThreadInfo()
		{
			super("threads");
		}
		
		public String[] getInfo()
		{
			String[] retval;
			java.lang.Thread[] allThreads;
			ThreadGroup r, p;

			// find root ThreadGroup
			r = moab_Thread.currentThread().jlThread.getThreadGroup();
			p = r;
			while (p != null)
			{
				r = p;
				p = r.getParent();
			}
			assert(r != null);

			allThreads = new java.lang.Thread[r.activeCount() + 50];
			int ct = r.enumerate(allThreads, true);
			
			retval = new String[ct];
			for (int i = 0; i < ct; i++)
			{
				retval[i] = allThreads[i].toString()
					+ " name="
					+ allThreads[i].getName();
			}

			return retval;
		}
	}

	static
	{
		Status.addInfoHook(new moab_ThreadInfo());
	}

	/**
	 * Create a new native thread.  This constructor is
	 * used to only create "explicit" threads.
	 *
	 * @allocations: A new ThreadWrapper. XXX.  Plus internal, native-thread state.
	 */
	PACKAGE moab_Thread(String name, int stackSize)
	{
		this.jlThread = new ThreadWrapper(name);
		
		//XXX this.threadPool = threadPool;

		DEBUGONLY(this.jlThread.setPriority(5);); // XXX XXX XXX

		TRACE(Trace.NODEOS_THREADS, " Created new thread: " +this);

		// XXX allocates a bunch of stuff
		setupNativeThreadState(this.jlThread, stackSize);
	}

	/**
	 * Setup most of the JVM-internal thread state.  This is a
	 * bit of of a hack to help us avoid hacking up the threading
	 * guts of the JVM into
	 */
	private native void setupNativeThreadState(java.lang.Thread jlthread,
						   int stackSize);

	PACKAGE static void yieldTo(moab_Thread th)
	{
		java.lang.Thread.yield();
	}

	PACKAGE static void sleep(long millis, int nanos)
		throws InterruptedException
	{
		java.lang.Thread.sleep(millis, nanos);
	}

	PACKAGE static Object getObject()
	{
		return self().obj;
	}

	PACKAGE static Object setObject(Object obj)
	{
		moab_Thread self = self();
		Object old;

		// XXX publically visible lock, but only hoses user
		synchronized(self) 
		{
			old = self.obj;
			self.obj = obj;
		}
		return old;
	}


	/**
	 * Package-private run method that all *implicit* threads
	 * start in.
	 */
	PACKAGE void start()
	{
		this.jlThread.start();
	}

	/**
	 * Return the Thread object associated with the currently
	 * executing thread of control.
	 *
	 * @see currentThread()
	 * @return a Thread object associated with the current thread.
	 */
	PACKAGE static moab_Thread self() 
	{
		return currentThread();
	}

	/**
	 *
	 * @see #self()
	 */
	PACKAGE static moab_Thread currentThread()
	{
		java.lang.Thread jth;
		jth = java.lang.Thread.currentThread();
#ifdef DEBUG
		if (!(jth instanceof ThreadWrapper))
		{
			Node.printStackTrace(null);
			Node.fatalError("ERROR: current thread is not a moab_Thread.ThreadWrapper!");
		}
#endif
		return ((ThreadWrapper)jth).moab_Thread();
	}

	public String toString()
	{
		String name = this.jlThread.getName();
		return "moab_Thread[n=" +name+ "]";
	}
}
