/*
 * moab_ThreadPool.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

/**
 * NodeOS ThreadPool implementation in Java.
 * <p>
 * Note that Moab does not actually expose its internal thread pool as an object.
 * This object just tracks the relevant thread objects and keeps them alive while
 * they are hanging out in Moab.  We cannot destroy any of the threads until
 * the entire thread pool is toasted, which requires the containing flow to die.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE final class moab_ThreadPool
{
	CHECK_NATIVE_NODEOS

	/**
	 * A list of the moab threads pre-allocated for this ThreadPool.
	 * This is here to ensure that the threads are *not* collected
	 * by the GC while they exist only in Moab.
	 */
	private final moab_Thread[] threads;

	/**
	 * Create an actual pool of explicit threads from the given
	 * specification.
	 */
	PACKAGE moab_ThreadPool(moab_ThreadPoolSpec spec)
	{
		this(spec, true);
	}
	
	/**
	 * Create an actual pool of threads from the given specification.
	 *
	 * @param spec The spec we should try to satisfy.
	 * @param explicit True if the pool should be made up of explicit
	 * @param explicit threads, false for implicit.  XXX Should this be
	 * @param explicit here?
	 */
	PACKAGE moab_ThreadPool(moab_ThreadPoolSpec spec, boolean explicit)
	{
		int rc;
		int maxThreads;
		int stackSize;
		int rate;

		require(spec);
		require(spec.maxThreads() >= 1);
		require((spec.stackSize() == 0)
			|| (spec.stackSize() >= 8*1024));

		// assert(spec.maxThreads() < 200); /* XXX just to be safe */

		maxThreads = spec.maxThreads();
		stackSize = spec.stackSize();
		rate = spec.getRate();

		if( explicit )
		{
			this.threads = new moab_Thread[maxThreads];
			
			/*
			 * Just *create* the threads, they are not started
			 * here.
			 */
			for (int i = 0; i < maxThreads; i++)
			{
				this.threads[i] =
					new moab_Thread(/*XXX noname*/null,
							stackSize);
			}
		}
		else
		{
			/* Don't bother making any. */
			this.threads = null;
		}
		
		/*
		 * Initialize native portion of this thread pool.
		 * XXX moab doesn't have exposed thread pool yet, so
		 * we're just initializing the underlying threadpoolspec.
		 */
		assert(spec.sched() == ThreadScheduler.basicScheduler());
		rc = init(maxThreads,
			  rate,
			  this.threads,
			  /* XXX spec: */ 0,
			  explicit);
		assert(rc == 0); // XXX
	}

	private native int init(int maxThreads,
				int rate,
				moab_Thread[] threads,
				int schedHack,
				boolean explicit);

	public String toString()
	{
		return "moab_ThreadPool["+threads.length+ " threads]";
	}

	PACKAGE void destroy()
	{
		if( this.threads != null )
		{
			// XXX drop threads on floor, let GC pick up.
			// we could be more agressive about destroying these
			for (int i = 0; i < threads.length; i++)
				threads[i] = null;
			
			// Cannot drop threads array, as it is final
		}
	}
}
