/*
 * moab_TransportInterface.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.lang.reflect.Constructor; // XXX 
import java.lang.reflect.InvocationTargetException; // XXX

/**
 * A moab_TransportInterface holds any important information about a network
 * interface in the system.
 */
PACKAGE class moab_TransportInterface
	implements TransportInterface
{
	/**
	 * The name of this transport, needs to match the string that will be
	 * given in any protocol specs.
	 */
	private String name;

	/**
	 * The, uh, logical id as reported by moab...
	 */
	private int id;

	/**
	 * The address of this interface, also used as a model when creating
	 * new addresses that pertain to this network.
	 */
	private AddressSpec address;

	/**
	 * The, uh, bandwidth as reported by moab...
	 */
	private int bandwidth;

	/**
	 * The transport is a real physical interface and not an upper level
	 * protocol.
	 */
	private boolean physical = false;
	
	private moab_TransportInterface()
	{
	}

	/**
	 * @param name The name of this interface, should match the name as
	 * @param name given in any protocol specs.
	 * @param id The logical id of this interface...
	 * @param address The address of this interface or an object that
	 * @param address new addresses should be modeled after.
	 * @param bandwidth The available bandwidth on this interface.
	 */
	PACKAGE moab_TransportInterface(String name, int id,
					AddressSpec address,
					int bandwidth)
	{
		this.name = name;
		this.id = id;
		this.address = address;
		this.bandwidth = bandwidth;
	}

	/**
	 * Construct an alias for another moab_TransportInterface.
	 *
	 * @param name The alias for the given interface.
	 * @param ti The interface to copy.
	 */
	PACKAGE moab_TransportInterface(String name,
					moab_TransportInterface ti)
	{
		this.name = name;
		this.id = ti.id;
		this.address = ti.address;
		this.bandwidth = ti.bandwidth;
	}

	/**
	 * @return The name of this interface.
	 */
	public String getName()
	{
		return this.name;
	}

	/**
	 * @return The id of this interface.
	 */
	public int getID()
	{
		return this.id;
	}

	/**
	 * @return The address of this interface.
	 */
	public AddressSpec getAddress()
	{
		return this.address;
	}

	/**
	 * @return The available bandwidth for this interface.
	 */
	public int getBandwidth()
	{
		return this.bandwidth;
	}

	/**
	 * @return Whether this is a physical interface or an upper level
	 * @return protocol transport.
	 */
	public boolean getPhysical()
	{
		return this.physical;
	}
	
	/**
	 * The parameters that AddressSpec objects will take.
	 */
	private static Class specParameters[] = { String.class };

	/**
	 * Create a new object that can Address something at this protocol
	 * level.
	 *
	 * @param spec The address specification string that should be specific
	 * @param spec to this protocol level.
	 */
	public AddressSpec createAddress(String spec)
	{
		AddressSpec retval = null;

		try
		{
			Object args[] = { spec };
			Constructor cons;
			Class cl;
			
			cl = this.address.getClass();
			cons = cl.getConstructor(specParameters);
			retval = (AddressSpec)cons.newInstance(args);
		}
		catch(NoSuchMethodException e)
		{
			/* XXX */
			Node.println("No such method error?");
		}
		catch(IllegalAccessException e)
		{
			/* XXX */
			Node.println("Illegal access?");
		}
		catch(InstantiationException e)
		{
			/* XXX */
			Node.println("InstantiationException");
		}
		catch(InvocationTargetException e)
		{
			/* XXX */
			Node.println("Invocation exception");
		}
		return retval;
	}

	public String toString()
	{
		return "moab_TransportInterface[name="
			+ this.name
			+ "; id="
			+ this.id
			+ "; address="
			+ this.address
			+ "; bandwidth="
			+ this.bandwidth
			+ "; physical="
			+ this.physical
			+ "]";
	}
}
