/*
 * pj_Flow.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Vector; // XXX

/**
 * Java implementation of a NodeOS Flow.
 * A flow is the closest approximation to a process in Janos.
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_Flow
{
	/**
	 * The credentials provided when this pj_Flow was created.  This
	 * is the bucket from which the pj_Flow gets to extract resources
	 * and access Node facilities.
	 */
	private final pj_Credentials cred;

	/**
	 * The threadPool for this flow.
	 */
	private final pj_ThreadPool threadPool;

	/**
	 * The Memory limit for this flow.
	 */
	private final pj_MemSpec memSpec;

	/**
	 * Channels created in this pj_Flow.  This list only exists
	 * so that the channels can be cleanly destroyed when the
	 * flow is terminated.
	 */
	private final Vector channels = new Vector(5);

	/**
	 * A per-flow Object that user code can set and get.
	 * (for associating things with the flow).
	 */
	private Object obj;

	/**
	 * XXX: need to examine the global locking structure for
	 * deadlock, etc
	 * <p>
	 * XXX: Probably should move to locks on each element of the
	 * state and not a global state lock.
	 */
	PACKAGE final Object lock = new Object();

	private final Runnable termHandler;
	private pj_Flow parentFlow = null;
	
	/**
	 * Create a new flow "under" the given parent.
	 *
	 * @see #createFlow()
	 * @see #createRootFlow()
	 */
	private static FlowHandle createFlowUnder(Credentials cred,
						  ThreadPoolSpec tPoolSpec,
						  MemSpec memSpec,
						  FlowStartSpec flowStartSpec,
						  Runnable termHandler,
						  pj_Flow parentFlow)
		throws FlowStartException
	{
		Flow f;

		require((parentFlow == null)
			|| (parentFlow == Flow.currentFlow().getNativeFlow()));

		try
		{
			/* Switch to heap of new flow */
			f = new Flow(cred, tPoolSpec, memSpec, termHandler,
				     FlowClassSpec.getDefault());
			/* Switch back to heap of parent flow */
		}
		catch (Throwable th)
		{
			throw new FlowStartException("huh?", th);
		}
		
		/* PJ-specific initialization */
		f.getNativeFlow().parentFlow = parentFlow; 

		/* XXX implicit thread start in a constructor is bad. */
		f.threadStart(new FlowInitializer(flowStartSpec),
			      "Initializer");
		
		return new FlowHandle(f);
	}

	private static final MemoryPolicy fakeMemPolicy = new MemoryPolicy(1000);

	PACKAGE MemoryPolicy getMemoryPolicy()
	{
		return fakeMemPolicy;
	}
	
	/**
	 * Odd method for this level (its just using the public interfaces).
	 * This is here because the moab_Flow.createFlow() is quite funky.
	 *
	 * @see Flow#newFlow()
	 */
	PACKAGE static FlowHandle createFlow(Credentials cred,
					     ThreadPoolSpec tPoolSpec,
					     MemSpec memSpec,
					     FlowStartSpec flowStartSpec,
					     Runnable termHandler)
		throws FlowStartException
	{
		return createFlowUnder(cred, tPoolSpec, memSpec,
				       flowStartSpec, termHandler,
				       Flow.currentFlow().getNativeFlow());
	}

	/**
	 * Creating root flow is no different from creating any other
	 * flow in pj version.  Much different in Moab version.
	 *
	 * @see Flow#newFlow()
	 */
	PACKAGE static FlowHandle createRootFlow(Credentials cred,
						 ThreadPoolSpec tPoolSpec,
						 MemSpec memSpec,
						 FlowStartSpec flowStartSpec)
		throws FlowStartException
	{
		FlowHandle fh;
		
		fh =  createFlowUnder(cred, tPoolSpec, memSpec,
				      flowStartSpec,
				      /* Root flow has no term handler: */
				      null,
				      /* Root flow has no parent: */
				      null);

		// XXX this is a hack:
		((pj_FlowHandle)fh).setName("Root Flow");

		return fh;
	}

	/**
	 * Initalize a new Flow.  See newFlow().
	 */
	PACKAGE pj_Flow(pj_Credentials cred,
			pj_ThreadPoolSpec threadPoolSpec,
			pj_MemSpec memSpec,
			Runnable termHandler)
	{
		this.cred = cred;
		this.memSpec = memSpec;

		this.termHandler = termHandler; // XXX

		this.threadPool = new pj_ThreadPool(this,
						    threadPoolSpec);
	}

	/**
	 * Create an explicit thread.  This is a new thread which is
	 * not subject to whatever concurrency limit this Flow's
	 * ThreadPool has.  But, it will be scheduled by the
	 * ThreadPool's scheduler.
	 * 
	 * @allocations: This function will allocate a new Thread object.
	 * @param r the Runnable to start the target thread on.
	 * @param name the String name of the new thread.
	 */
	PACKAGE void threadStart(java.lang.Runnable r, String name)
	{
		this.threadPool.threadStart(r, name);
	}

	private boolean destroyed = false;

	PACKAGE synchronized void destroy()
	{
		// For debugging, make sure this isn't invoked more than once
		assert(!destroyed);
		this.destroyed = true;

		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		
		// Run the termination handler in the parent flow (if this isn't the root)
		if (this.parentFlow != null)
		{
			TRACE(Trace.NODEOS_DESTROY, "Kicking parent to run term handler: " +this.termHandler);
			assert(this.termHandler);
			parentFlow.threadStart(this.termHandler, "Termination Handler");
		}
		else
		{
			// then this is the the root flow
			assert(this.termHandler == null);

			Node.shutdown();
		}

		for (int i = 0; i < channels.size(); i++)
		{
			// Yes, this is stupid.  It is just temporary.
			// XXX make all the pj_channels extend
			// abstract class pj_AnyChannel

			Object c = channels.elementAt(i);
			if (c instanceof pj_InChannel)
			{
				((pj_InChannel)c).destroy();
			}
			else if (c instanceof pj_OutChannel)
			{
				((pj_OutChannel)c).destroy();
			}
			else if (c instanceof pj_CutThroughChannel)
			{
				((pj_CutThroughChannel)c).destroy();
			}
			else
			{
				panic("Non-channel object in channel list.");
			}
		}

		TRACE(Trace.NODEOS_DESTROY, "Destroying thread pool");
		threadPool.destroy();
	}
	
	PACKAGE void addChannel(pj_InChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE void addChannel(pj_OutChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE void addChannel(pj_CutThroughChannel c)
	{
		channels.addElement(c);
	}

	PACKAGE Object getObject()
	{
		return obj;
	}

	PACKAGE void setObject(Object obj)
	{
		this.obj = obj;
	}

	PACKAGE pj_ThreadPool getThreadPool()
	{
		return threadPool;
	}

	PACKAGE void getStats(pj_FlowStats fs)
	{
		// XXX no stats in pj, yet
	}

	/**
	 * Return a fragment of an id string.  Flow.toString() will wrap this
	 */
	public String toString()
	{
		if (false) // longwinded
			return  "cred=" + cred
				+ "; obj=" + obj
				+ "; objid=0x" + Integer.toHexString(this.hashCode());
		else // short and sweet
			return "objid=0x" + Integer.toHexString(this.hashCode());
	}

	/**
	 * Get the current flow.  The current Flow is the
	 * flow associated with the current Thread.
	 *
	 * @see Thread#getFlow()
	 */
	PACKAGE static pj_Flow native_currentFlow()
	{
		return pj_Thread.self().getFlow();
	}
}

