/*
 * pj_InChannelDispatcher.Java
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

import java.util.Hashtable;
import java.util.Vector;
import java.io.IOException;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Class that is used to dispatch packets from protocol layers to 
 * the appropriate InChannel.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE class pj_InChannelDispatcher
	extends pj_ProtocolLayer.PacketHandler
{
	/**
	 * The set of dispatchers for the system.
	 * <p>
	 * XXX global state?
	 */
	private static final Hashtable dispatchers = new Hashtable();

	/**
	 * Get the dispatcher that matches the given source mask.
	 *
	 * @param pl The protocol layer were packets are coming from
	 * @param source The source address mask packets are matched too
	 */
	PACKAGE static pj_InChannelDispatcher
	getDispatcher(pj_ProtocolLayer pl,
		      pj_ProtocolLayer.PacketAddressMask source)
		throws IOException
	{
		pj_InChannelDispatcher dispatcher;

		synchronized (pj_InChannelDispatcher.dispatchers)
		{
			/*
			 * Search our hash table first, and then mask one if
			 * needed
			 */
			dispatcher = (pj_InChannelDispatcher)
				pj_InChannelDispatcher.dispatchers.get(source);

			if (dispatcher == null)
			{
				dispatcher = new pj_InChannelDispatcher();
				dispatcher.protocolLayer = pl;
				dispatcher.source = source;
				pj_InChannelDispatcher.dispatchers.put(source, dispatcher);
				pl.attachHandler(dispatcher);
			}
		}
		return dispatcher;
	}

	/**
	 * The protocol layer that this dispatcher is attached to
	 */
	private pj_ProtocolLayer protocolLayer;
	
	/**
	 * The set of channels that this handler can dispatch too.
	 * We only really care about their demultiplex keys since
	 * the rest of the demultiplex is done elsewhere.
	 */
	private final Vector channels = new Vector();

	/**
	 * Add a channel to this dispatcher.  The channel is placed
	 * in the selection line by how large its key is, the larger
	 * the key the earlier it will be checked.
	 */
	PACKAGE void addChannel(pj_InChannel newChan)
	{
		int lpc, size;

		synchronized(this.channels)
		{
			size = this.channels.size();
			for( lpc = 0; lpc < size; lpc++ )
			{
				pj_InChannel elem;

				elem = (pj_InChannel)this.channels.elementAt(lpc);

				/* Insert in biggest to smallest order */
				if( newChan.classifier.length() >=
				    elem.classifier.length() )
				{
					break;
				}
			}
			TRACE(Trace.NODEOS_DISPATCH, "Adding packet classifier " + newChan.classifier);
			this.channels.insertElementAt(newChan, lpc);
		}
	}

	/**
	 * Remove a channel from this dispatcher.
	 */
	PACKAGE void remChannel(pj_InChannel deadChan)
	{
		synchronized(this.channels)
		{
			this.channels.removeElement(deadChan);
			if( this.channels.size() == 0 )
			{
				/*
				 * No more channels, detach ourselves from the
				 * protocol layer
				 */
				this.protocolLayer.detachHandler(this);
				synchronized(pj_InChannelDispatcher.dispatchers)
				{
					pj_InChannelDispatcher.dispatchers.
						remove(this.source);
				}
			}
		}
	}

	/**
	 * Handle a packet that came in through the transport layer.
	 * Returns true if the packet was handled.
	 * <p>
	 * The buffer is handled only if an InChannel can be found and
	 * that InChannel has buffers available to swap with.
	 * 
	 * @context Operating in "interrupt context".  (no real interrupts, mind you). 
	 * @param pb The system's receive pj_BufferHandle
	 * @return the input pb, perhaps with a different underlying buffer
	 */
	PACKAGE boolean handle(pj_BufferHandle pb)
	{
		pj_InChannel matchChan = null;
		int payloadStart = 0;
		int lpc, size;

		/**
		 * Walk all the channels attached to this handler
		 * and find a match.
		 */
		synchronized(this.channels)
		{
			size = this.channels.size();
			for( lpc = 0;
			     (lpc < size) && (matchChan == null);
			     lpc++ )
			{
				InChannel elem = (InChannel)this.channels.
					elementAt(lpc);

				payloadStart = elem.classifier.
					matches(pb.getByteArray(), 0,
						pb.getLength());
				if( payloadStart != -1 )
				{
					/* Found a match */
					matchChan = elem;
				}
			}
		}
		if( matchChan != null )
		{
			TRACE(Trace.NODEOS_INCOMING, "demultiplexed to: " +matchChan.classifier);

			/*
			 * Ask the channel to take the matching buffer
			 * and return a replacement.  Channel will
			 * return false if it has nothing to swap for
			 * it.
			 */
			if (matchChan.handle(pb, 0, payloadStart))
				return true;

			/*
			 * The channel can't afford the packet.
			 */
			TRACE(Trace.NODEOS_INCOMING, "  Channel can't afford packet.");
		}
		else
		{
			TRACE(Trace.NODEOS_INCOMING, "No matching channel for pkt: data=" +Debug.dbgPacketSummary(pb));
		}

		return false;
	}
}
