/*
 * pj_ThreadPool.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_NODEOS_PKG;

/**
 * NodeOS ThreadPool implementation in Java.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
PACKAGE final class pj_ThreadPool
	implements Runnable
{
	/**
	 * Lock protecting the queue of WorkItems.
	 */
	private final Object bufferQLock;

	/**
	 * Queued buffers (waiting for a thread to handle them) are stored here.
	 */
	private pj_BufferHandle bufferQueue;

	/**
	 * The flow this ThreadPool is associated with.
	 */
	private final pj_Flow flow;

	private boolean destroyed;

	private final String unassignedName = "Unassigned worker";

	/**
	 * Create a new pj_ThreadPool.
	 * <p>
	 * The variousSchedulerParameters and scheduler parameters
	 * hardly even qualify as hacks.
	 */
	PACKAGE pj_ThreadPool(pj_Flow flow, pj_ThreadPoolSpec tpspec)
	{
		require(flow);
		require(tpspec);

		this.bufferQLock = new Object();
		this.destroyed = false;
		this.flow = flow;

		for (int i = 0; i < tpspec.maxThreads(); i++)
		{
			pj_Thread th;
			th = new pj_Thread(this, this, unassignedName);
			th.start();
		}
	}

	/**
	 * Create an explicit thread and return that thread handle.
	 * This is a new thread which is not subject to whatever
	 * concurrency limit the pj_ThreadPool has.  But, it will be
	 * scheduled by the pj_ThreadPool's scheduler.
	 * 
	 * @allocations: This function will allocate a new Thread object.
	 * @param r the Runnable to start the target thread on.
	 */
	PACKAGE void threadStart(java.lang.Runnable runnable, String name)
	{
		require(runnable);

		if (destroyed())  // XXX
			return;

		pj_Thread th = new pj_Thread(this, runnable, name);
		th.start();
	}

	/**
	 * Return the pj_Flow this ThreadPool is associated with.
	 */
	PACKAGE pj_Flow native_getFlow()
	{
		return this.flow;
	}

	/**
	 * Destroy this pj_ThreadPool.
	 * <p>
	 * XXX this is a hack.
	 */
	PACKAGE void destroy()
	{
		TRACE(Trace.NODEOS_DESTROY, "Destroying " + this);
		synchronized(bufferQLock)
		{
			destroyed = true;
			bufferQLock.notifyAll();
		}
		TRACE(Trace.NODEOS_DESTROY, "  Marked as destroyed (" +this+ ")");
	}

	PACKAGE boolean destroyed()
	{
		return destroyed;
	}


	/**
	 * Add a Buffer to the pj_ThreadPool's Buffer list.
	 * <p>
	 * The buffer must not be on some other list (either
	 * an InChannel list, OutChannel list, or some bufferQueue).
	 */
	PACKAGE void addWork(pj_BufferHandle bh)
	{
		require(bh.notOnAQueue());

		synchronized(bufferQLock)
		{
			bufferQueue = bh.addToQueueHead(bufferQueue);
			signalWorkers();
		}
	}

	PACKAGE pj_BufferHandle dequeueWorkItem()
	{
		pj_BufferHandle bh = null;

		assertLocked(this.bufferQLock);

		/*
		 * Take from the tail of the queue.  In a singleton
		 * list, that means take the head and set head to null.
		 * Otherwise, just remove bufferQueue.prev.
		 */

		if (this.bufferQueue != null)
		{
			// Take prev from head (last element)
			bh = this.bufferQueue.prev();
			assert(bh != null);

			// Remove it from the queue.
			bh.removeFromQueue();

			// If queue would be empty, null head
			if (bh == this.bufferQueue)
				this.bufferQueue = null;
		}

		return bh;
	}
	
	PACKAGE void signalWorkers()
	{
		assertLocked(bufferQLock);
		bufferQLock.notify(); // XXX notifyAll() ???
	}

	/**
	 * Do the work on this pj_ThreadPool.
	 * <p>
	 * This function won't return until the ThreadPool is destroyed.
	 */
	public void run()
	{
		pj_BufferHandle bh;

		TRACE(Trace.NODEOS_THREADS, " Worker thread started work loop.");

		while(!destroyed())
		{
			bh = null;

			synchronized(bufferQLock)
			{
				try
				{
					bh = dequeueWorkItem();
					if ((bh == null)
					    && (!destroyed()))
						bufferQLock.wait();
				}
				catch (java.lang.InterruptedException ie)
				{
					// from the wait().  Just ignore it.
				}
			}

			// Do the item outside the context of the bufferQLock!
			if (bh != null)
			{
				DeliverFunc receiver;
				int payloadStart;

				TRACE(Trace.NODEOS_THREADS, " Worker thread started on " +bh);
				DEBUGONLY(pj_Thread.self().setName("Working on " +bh));

				receiver = bh.getDeliverFunc();
				payloadStart = bh.getPayloadStart();

				DEBUGONLY(bh.clearIncomingState());

				require(bh instanceof BufferHandle);

				try
				{
					((BufferHandle)bh).userOwns();
					receiver.receive((BufferHandle)bh, payloadStart);
				}
				catch(Throwable th)
				{
					Node.println("Uncaught exception in deliverFunc (" +receiver+ "):");
					Node.println(th.getMessage());
					Node.printStackTrace(th);
				}

				DEBUGONLY(pj_Thread.self().setName(unassignedName));

				// Cannot do anything with the
				// pj_BufferHandle at this point, it
				// might already be scheduled for
				// re-use.
				bh = null;
			}
		}
		
		TRACE(Trace.NODEOS_DESTROY, " Worker Thread " +pj_Thread.self()+ " destroyed.");
	}
}
