/*
 * ICMPEcho.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_PROTOCOLS_PKG.icmp;

import JANOS_NODEOS_PKG.BufferHandle;

import JANOS_PROTOCOLS_PKG.ip.IPHeader;

import JANOS_UTIL_PKG.Packet;
import JANOS_UTIL_PKG.PacketSegment;
import JANOS_UTIL_PKG.PlainPacketSegment;

/**
 * An ICMPEcho is a full ICMP ECHO_REQUEST packet, including the ip header,
 * ICMP header, and a body segment.
 */
public class ICMPEcho
	extends Packet
{
	/**
	 * The index of the IP header in the array of PacketSegments.
	 */
	public static final int IP_SLOT		= 0;
	
	/**
	 * The index of the ICMP header in the array of PacketSegments.
	 */
	public static final int ICMP_SLOT	= 1;
	
	/**
	 * The index of the Identifier/SequenceNumber option in the array of
	 * PacketSegments.
	 */
	public static final int ID_SEQ_SLOT	= 2;

	/**
	 * Construct an ICMP ECHO_REQUEST packet.
	 *
	 * @param src The source address for the packet.
	 * @param dst The destination address for the packet.
	 * @param body The body of the echo that will be sent back in the
	 * @param body reply packet.
	 */
	public ICMPEcho(int src, int dst, PacketSegment body)
	{
		super(new PacketSegment[] {
			new IPHeader((byte)0,
				     (short)0,
				     (short)0,
				     IPHeader.MAX_TTL,
				     IPHeader.PROTO_ICMP,
				     src,
				     dst),
			new ICMPHeader(ICMPHeader.ECHO, (byte)1),
			body
		});
	}
	
	/**
	 * Construct an ICMP ECHO_REQUEST packet with the optional identifier
	 * and sequence number fields.
	 *
	 * @param src The source address for the packet.
	 * @param dst The destination address for the packet.
	 * @param identifier The value for the identifier field in the packet.
	 * @param sequenceNumber The value of the sequence number field in the
	 * @param sequenceNumber packet.
	 * @param body The body of the echo that will be sent back in the
	 * @param body reply packet.
	 */
	public ICMPEcho(int src, int dst,
			short identifier, short sequenceNumber,
			PacketSegment body)
	{
		super(new PacketSegment[] {
			new IPHeader((byte)0,
				     (short)0,
				     (short)0,
				     IPHeader.MAX_TTL,
				     IPHeader.PROTO_ICMP,
				     src,
				     dst),
			new ICMPHeader(ICMPHeader.ECHO, (byte)0),
			new PlainPacketSegment(new byte[] {
				(byte)((identifier >> 8) & 0xff),
				(byte)((identifier     ) & 0xff),
				(byte)((sequenceNumber >> 8) & 0xff),
				(byte)((sequenceNumber     ) & 0xff)
			}),
			body
		});
	}

	/**
	 * Construct, but don't initialize, an ICMP ECHO_REQUEST packet.
	 *
	 * @param body The body of the echo packet.
	 */
	public ICMPEcho(PacketSegment body)
	{
		super(new PacketSegment[] {
			new IPHeader(),
			new ICMPHeader(),
			body
		});
	}
	
	/**
	 * Construct, but don't initialize, an ICMP ECHO_REQUEST packet with
	 * the identifier and sequence number optional fields.
	 */
	public ICMPEcho(boolean withID, PacketSegment body)
	{
		super(new PacketSegment[] {
			new IPHeader(),
			new ICMPHeader(),
			new PlainPacketSegment(4),
			body
		});
	}

	/**
	 * @return The application specific identifier value.
	 */
	public short getIdentifier()
	{
		return ((PlainPacketSegment)this.getSegments()[ID_SEQ_SLOT]).
			getShort(0);
	}

	/**
	 * @param id The application specific identifier value.
	 */
	public void setIdentifier(short id)
	{
		((PlainPacketSegment)this.getSegments()[ID_SEQ_SLOT]).
			setShort(0, id);
	}

	/**
	 * @return The application specific sequence number value.
	 */
	public short getSequenceNumber()
	{
		return ((PlainPacketSegment)this.getSegments()[ID_SEQ_SLOT]).
			getShort(2);
	}
	
	/**
	 * @param seqNum The application specific sequence number value.
	 */
	public void setSequenceNumber(short value)
	{
		((PlainPacketSegment)this.getSegments()[ID_SEQ_SLOT]).
			setShort(2, value);
	}
}
