/*
 * ICMPHeader.Java
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_PROTOCOLS_PKG.icmp;

import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.DemultiplexKey;

import JANOS_UTIL_PKG.PacketSegment;
import JANOS_UTIL_PKG.InvalidPacketException;

/**
 * The ICMPHeader is a PacketSegment for the standard ICMP header.
 */
public class ICMPHeader
	implements PacketSegment
{
	/**
	 * Message types and their codes.
	 */
	public static final byte ECHOREPLY	=  0;
	public static final byte UNREACH	=  3;
	public static final byte	UNREACH_NET			=  0;
	public static final byte	UNREACH_HOST			=  1;
	public static final byte	UNREACH_PROTOCOL		=  2;
	public static final byte	UNREACH_PORT			=  3;
	public static final byte	UNREACH_NEEDFRAG		=  4;
	public static final byte	UNREACH_SRCFAIL			=  5;
	public static final byte	UNREACH_NET_UNKNOWN		=  6;
	public static final byte	UNREACH_HOST_UNKNOWN		=  7;
	public static final byte	UNREACH_ISOLATED		=  8;
	public static final byte	UNREACH_NET_PROHIB		=  9;
	public static final byte	UNREACH_HOST_PROHIB		= 10;
	public static final byte	UNREACH_TOSNET			= 11;
	public static final byte	UNREACH_TOSHOST			= 12;
	public static final byte	UNREACH_FILTER_PROHIB		= 13;
	public static final byte	UNREACH_HOST_PRECEDENCE		= 14;
	public static final byte	UNREACH_PRECEDENCE_CUTOFF	= 15;
	public static final byte SOURCEQUENCH	=  4;
	public static final byte REDIRECT	=  5;
	public static final byte	REDIRECT_NET		= 0;
	public static final byte	REDIRECT_HOST		= 1;
	public static final byte	REDIRECT_TOSNET		= 2;
	public static final byte	REDIRECT_TOSHOST	= 3;
	public static final byte ECHO		=  8;
	public static final byte ROUTERADVERT	=  9;
	public static final byte ROUTERSOLICIT	= 10;
	public static final byte TIMXCEED	= 11;
	public static final byte	TIMXCEED_INTRANS	= 0;
	public static final byte	TIMXCEED_REASS		= 1;
	public static final byte PARAMPROB	= 12;
	public static final byte	PARAMPROB_OPTABSENT	= 1;
	public static final byte TSTAMP		= 13;
	public static final byte TSTAMPREPLY	= 14;
	public static final byte IREQ		= 15;
	public static final byte IREQREPLY	= 16;
	public static final byte MASKREQ	= 17;
	public static final byte MASKREPLY	= 18;
	public static final byte MAXTYPE	= 18;

	/**
	 * Offsets into an ICMP header in a buffer.
	 */
	private static final int OFFSET_TYPE	= 0;
	private static final int LENGTH_TYPE	= 1;
	private static final int OFFSET_CODE	= 1;
	private static final int LENGTH_CODE	= 1;
	private static final int OFFSET_CHKSUM	= 2;
	private static final int LENGTH_CHKSUM	= 2;

	/**
	 * Length of the header.
	 */
	public static final int HEADER_LENGTH	= 4;

	/**
	 * ICMP header data.
	 */
	private byte type;
	private byte code;

	/**
	 * @param dk The key to add the matching data to.
	 * @param offset Offset of the IP header in the buffer.
	 * @param type The ICMP message type to match.
	 */
	public static void matchType(DemultiplexKey dk, int offset, byte type)
	{
		final byte seq[] = { type };
		
		dk.add(offset + OFFSET_TYPE, seq.length, seq, null);
	}

	/**
	 * @param dk The key to add the matching data to.
	 * @param offset Offset of the IP header in the buffer.
	 * @param code The ICMP message code to match.
	 */
	public static void matchCode(DemultiplexKey dk, int offset, byte code)
	{
		final byte seq[] = { code };
		
		dk.add(offset + OFFSET_CODE, seq.length, seq, null);
	}

	/**
	 * @param type The type to give this header segment.
	 * @param code The code to give this header segment.
	 */
	public ICMPHeader(byte type, byte code)
	{
		this.type = type;
		this.code = code;
	}

	/**
	 * @param type The type to give this header segment.
	 */
	public ICMPHeader(byte type)
	{
		this(type, (byte)0);
	}
	
	/**
	 * Construct an ICMPHeader with zero type and code values.
	 */
	public ICMPHeader()
	{
	}

	public int getLength()
	{
		return HEADER_LENGTH;
	}

	public byte getType()
	{
		return this.type;
	}

	public void setType(byte type)
	{
		this.type = type;
	}

	public byte getCode()
	{
		return this.code;
	}

	public void setCode(byte code)
	{
		this.code = code;
	}

	public int marshal(BufferHandle bh, int offset)
	{
		short checksum;
		
		bh.setByte(offset + OFFSET_TYPE, this.type);
		bh.setByte(offset + OFFSET_CODE, this.code);
		bh.setShort(offset + OFFSET_CHKSUM, (short)0);
		checksum = bh.computeChecksum(offset, bh.getLength() - offset);
		bh.setShort(offset + OFFSET_CHKSUM, checksum);
		return offset + HEADER_LENGTH;
	}

	public int unmarshal(BufferHandle bh, int offset)
		throws InvalidPacketException
	{
		if( bh.computeChecksum(offset, bh.getLength() - offset) != 0 )
			throw new InvalidPacketException("Bad Checksum");
		this.type = bh.getByte(offset + OFFSET_TYPE);
		this.code = bh.getByte(offset + OFFSET_CODE);
		return offset + HEADER_LENGTH;
	}

	public int hashCode()
	{
		return ((this.type << 8) & 0xff) |
			(this.code & 0xff);
	}
	
	public boolean equals(Object obj)
	{
		boolean retval = false;

		if( obj instanceof ICMPHeader )
		{
			ICMPHeader icmp = (ICMPHeader)obj;

			if( this.type == icmp.type
			    && this.code == icmp.code )
			{
				retval = true;
			}
		}
		return retval;
	}
	
	public String toString()
	{
		return "[ICMPHeader type="
			+ this.type
			+ "; code="
			+ this.code
			+ "]";
	}
}
