/* 
 * RouteEntry.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 *  This file is part of the Janos Java NodeOS software.
 *  
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation is hereby granted without fee, provided that the above
 *  copyright notice and this permission/disclaimer notice is retained in all
 *  copies or modified versions, and that both notices appear in supporting
 *  documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 *  WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 *  ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *  
 *  Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 *  improvements that they make and grant Univ. of Utah redistribution rights.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ROUTE_PKG;

import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.AddressFormatException;
import JANOS_NODEOS_PKG.NoSuchNodeAddress;
IMPORT_DEBUG

/**
 * A single entry of route table.  A route entry contains both
 * the "logical" address and "physical" address of a Node.
 * <p>
 * A RouteEntry is immutable.  That way code can read the
 * table but be trusted to not mess with it.  Note that this
 * immutablity depends on Address also being immutable.
 * <p>
 * XXX rename "logical" to "active" in all names.
 * XXX fixup names of local variables.
 *
 * @see RouteTable
 * @author David Wetherall
 */
public class RouteEntry
	implements Cloneable
{
	/**
	 * Destination address this route provides information
	 * for.
	 * <p>
	 * XXX Store an an Integer
	 */
	private final int dst; 

	/**
	 * Next hop address for a route.  This is an active
	 * address (vs. the physical address).
	 * <p>
	 * XXX Store as an Integer
	 * <p>
	 * XXX redundant with addr.getANAddress()?  Except
	 * when addr.getANAddress == 0.... blech.
	 */
	private final int nxt; // logical next address

	/**
	 * Next hop address for a route.  This is a physical address
	 * of the next hop.
	 */
	private Address addr;

	/**
	 * Create a new RouteEntry which is a copy of the given RouteEntry.
	 * This creates a shallow copy (both point to the same physical
	 * address object).
	 */
	public RouteEntry(RouteEntry other)
	{
		// require(other);
		this(other.dst, other.nxt, other.addr);
	}
	
	/**
	 * Create a new RouteEntry for the given destination
	 * whose next hop is the given nxt/addr.
	 * <p>
	 * XXX 'nxt' is redundant with the addr.getANAddress().
	 *
	 * @param dst the destination for the route, as an active address
	 * @param nxt the next hop for the route, as an active address
	 * @param addr the next hop for the route, as a physical address
	 */
	public RouteEntry(int dst, int nxt, Address addr)
	{
		require(dst != 0);
		require(nxt != 0);
		require(addr);
		require((addr.getANAddress() == nxt)
			|| (addr.getANAddress() == 0));

		this.dst = dst;
		this.nxt = nxt;
		this.addr = addr;
	}

	/**
	 * Create a new RouteEntry for the given destination
	 * whose next hop is the given next/addr.  The physical
	 * address must be a string of the form
	 * "<var>host</var>[:<var>port</var>[:<var>transport</var>]]"
	 * <p>
	 * XXX port default to 4242 if not provided?
	 * <p>
	 * XXX transport defaults to 'udp' if not provided.  'if/ipv4/' is
	 * prefixed on transport, regardless.  Which is bogus.
	 * <p>
	 * XXX why is there string address parsing code here?  Why not
	 * use the code in Address??
	 *
	 * @param d destin
	 */
	public RouteEntry(int dst, int nxt, String a)
		throws NumberFormatException, AddressFormatException, NoSuchNodeAddress
	{
		int port;
		String host;
		String transport = "udp";

		require(dst != 0);
		require(nxt != 0);
		require(a);

		this.dst = dst;
		this.nxt = nxt;
		
		// Parse the physical next hop address

		
		// Split a into host:port
		int sep = a.indexOf(':');
		if (sep == 0)
		{
			host = a;
			port = 4242; // XXXX default port
		}
		else
		{
			int tsep = a.indexOf(':', sep + 1);
			
			try
			{
				host = a.substring(0,sep);
				port = Integer.parseInt(a.substring(sep+1,
								    (tsep > 0) ?
								    tsep :
								    a.length()));
				if( tsep > 0 )
					transport = a.substring(tsep+1,
								a.length());
				// above might throw NumberFormatException
			}
			catch (StringIndexOutOfBoundsException oobe)
			{
				addr = null;
				panic("RouteEntry: parsing problem: " +oobe.getMessage()); // XXX how about something graceful?
			}
		}
		
		addr = AddressFactory.newAddress(nxt,
						 "if/ipv4/" +transport,
						 "*:" +port+ ":" +host);
	}
	
	public final int getLogicalDest()
	{
		return dst;
	}

	public final int getLogicalNext()
	{
		return nxt;
	}

	public final Address getPhysicalNext()
	{
		return addr;
		
	}

	public Object clone()
	{
		Object retval = null;
		
		try
		{
			RouteEntry re;
			
			retval = super.clone();
			re = (RouteEntry)retval;
			re.addr = (Address)this.addr.clone();
		}
		catch(CloneNotSupportedException cnse)
		{
		}
		return retval;
	}

	public int hashCode()
	{
		return this.dst;
	}
	
	public boolean equals(Object obj)
	{
		boolean retval = false;
		
		if( obj instanceof RouteEntry )
		{
			RouteEntry re = (RouteEntry)obj;

			if( re.dst == this.dst &&
			    re.nxt == this.nxt &&
			    re.addr.equals(this.addr) )
			{
				retval = true;
			}
		}
		return retval;
	}
	
	public String toString() {
		return "RouteEntry[" +
			"d=" +Address.toString(dst)+ ";" +
			// "n=" +Address.toString(nxt)+ ";" +
			"a=" +addr+
			"]";
	}
}
