/* 
 * RouteTable.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_ROUTE_PKG;

import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.AddressFormatException;
import JANOS_NODEOS_PKG.NoSuchNodeAddress;

import JANOS_NODEOS_PKG.Status;
import JANOS_NODEOS_PKG.InfoHook;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

import JVM_SYS_PKG.Visitor;
import JVM_SYS_PKG.LocalTeam;
import JVM_SYS_PKG.Importable;
import JVM_SYS_PKG.DeadTeamException;
import JVM_SYS_PKG.ThrowInVisitException;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Routing table.  Maps 32-bit integer "active addresses" to
 * RouteEntries.  A RouteEntry contains the core Address object
 * which, in turn, contains the "real" address of the target node.
 * RouteEntry can be subclassed to add more information to the table.
 * <p>
 * XXX most the hashtable used in this class is indexed by 'ints'.
 * But, that requires allocating (and dropping) a lot of "Integer"
 * objects.  Could be optimized by moving to a more pure hashtable (or
 * putting the Integer in the RouteTable.  Also, would benefit from
 * parameterized tables.
 * <p>
 * By design, this class defines a read-only RouteTable for users
 * in other packages.  Access to WriteableRouteTable is required
 * to make updates to the RouteTable.
 * <p>
 * Note that the read-only-ness of this class depends on the
 * immutablity of base RouteEntries.
 *
 * @see WriteableRouteTable
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public class RouteTable
	extends Importable
{
	/**
	 * Construct a table for routes.  The initial table is empty.
	 * Routes can be added through <code>addRoute()</code> or
	 * ...
	 * <p>
	 * XXX take a size hint for the hashtable.
	 * 
	 */
	public RouteTable()
	{
		LocalTeam.fetch().getKernelTeam().
			importNewObject(this, RemoteRouteTable.class);
	}
	
	public synchronized Enumeration getRoutes()
	{
		return ((RemoteRouteTable)this.get()).routes.elements();
	}

	public synchronized Enumeration getNeighbors()
	{
		return ((RemoteRouteTable)this.get()).neighbors.elements();
	}

	public synchronized int routeCount()
	{
		return ((RemoteRouteTable)this.get()).routes.size();
	}

	public synchronized int neighborCount()
	{
		return ((RemoteRouteTable)this.get()).neighbors.size();
	}

	/**
	 * A convenience function for casting objects
	 * returned from the routes hashtable.
	 * <p>
	 * XXX Using parameterized Hashtables would eliminate this,
	 * methinks.
	 * XXX can't this assert() that all o's are instances
	 * of RouteEntry or null?
	 */
	private static RouteEntry castToRouteEntry(Object o)
	{
		if (o instanceof RouteEntry)
		{
			return (RouteEntry)o;
		}
		return null;
	}
	
	/**
	 * Retrive the current route to get to the given active address.
	 */
	public synchronized RouteEntry getRoute(int activeDest)
	{
		return ((RemoteRouteTable)this.get()).getRoute(activeDest);
	}

	public synchronized RouteEntry getRoute(Integer activeDest)
	{
		return ((RemoteRouteTable)this.get()).getRoute(activeDest);
	}

	/**
	 * Remove a route from the table.
	 * <p>
	 * External access through WriteableRouteTable
	 */
	PACKAGE synchronized void package_removeRoute(RouteEntry r)
	{
		((RemoteRouteTable)this.get()).package_removeRoute(r);
	}

	/**
	 * Add a route to the table.
	 * <p>
	 * External access through WriteableRouteTable
	 */
	PACKAGE synchronized void package_addRoute(final RouteEntry route) 
	{
		final RemoteRouteTable rrt = (RemoteRouteTable)this.get();
		
		try
		{
			rrt.visitOwner(new Visitor() {
				public void run()
				{
					rrt.package_addRoute((RouteEntry)route.
							     clone());
				}
			});
		}
		catch(DeadTeamException dte)
		{
		}
		catch(ThrowInVisitException tive)
		{
		}
	}
}
