/*
 * Tester.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_TESTS_PKG;

import JANOS_NODEOS_PKG.Node;
import JANOS_NODEOS_PKG.Flow;

/**
 * The Tester class is used as a base for a class which does testing.
 */
abstract public class Tester
{
	/**
	 * Flags that are used to encode the build environment
	 */
	public static final int ENV_NONE = 0x0000;
	public static final int ENV_DEBUG = 0x0001;
	public static final int ENV_NDEBUG = 0x0002;
	public static final int ENV_EXPOSED = 0x0004;

	/**
	 * String names for the environment identifiers
	 */
	public static final String ENV_NAMES[] =
	{
		"debug",
		"ndebug",
		"exposed"
	};

	/**
	 * Generate verbose output
	 */
	PACKAGE static boolean verbose = false;

	/**
	 * The name of the current section we're testing
	 */
	private String section;

	/**
	 * Verifies that the class is running in the required environment
	 * and prints out a warning if it isn't.
	 *
	 * @param required Bitmask indicating the required environment
	 * @param mesg The section that is being tested
	 * @return True if the required environment is available, false
	 * otherwise
	 */
	private boolean verifyEnvironment(int required, String mesg)
	{
		boolean retval;

		if( (TestMain.getEnvironment() & required) == required )
		{
			retval = true;
		}
		else
		{
			int lpc;

			Node.println("Warning:  Cannot test section "
				     + mesg
				     + " in "
				     + this.getClass().getName()
				     + " because the class requires a"
				     + " different environment from the"
				     + " current build.");
			Node.print("The required environment(s) are:");
			for( lpc = 0; lpc < Tester.ENV_NAMES.length; lpc++ )
			{
				if( (required & (1 << lpc)) != 0 )
					Node.print(" " + Tester.
						   ENV_NAMES[lpc]);
			}
			Node.println("");
			retval = false;
		}
		return( retval );
	}

	/**
	 * Used to indicate that the tester class is entering a new section
	 * of testing.
	 *
	 * @param environment The required environment for this section
	 * @param desc A string naming/describing this section
	 * @return True if the requirements for the environment were met
	 */
	protected boolean enterSection(int environment, String desc)
	{
		boolean retval;

		retval = this.verifyEnvironment(environment, desc);
		if( retval )
		{
			this.section = desc;
			if( Tester.verbose )
				Node.println("Entering section: "
					     + desc);
		}
		return retval;
	}

	/**
	 * Get description of the current section
	 */
	public String getSection()
	{
		return this.section;
	}

	/**
	 * Test a boolean value and throw a TestFailedException if it was false
	 */
	protected void test(boolean value) throws TestFailedException
	{
		if( !value )
			throw new TestFailedException(this.section);
	}

	/**
	 * Test a boolean value and throw a TestFailedException if it was false
	 */
	protected void test(String name, boolean value) throws TestFailedException
	{
		if( !value )
			throw new TestFailedException(name + ": " + this.section);
	}

	/**
	 * Run the actual tests
	 */
	abstract public void testUnit() throws TestFailedException;

	/**
	 * Test the given class object
	 *
	 * @param cl A subclass of Tester that implements testUnit
	 */
	public static void testClass(Class cl)
	{
		Tester testable;

		try
		{
			testable = (Tester)cl.newInstance();
			Node.println("Now testing "
				     + cl.getName());
			testable.testUnit();
			Node.println("The test succeeded!");
		}
		catch(TestFailedException e)
		{
			Node.println("The test failed in section "
				     + e.getMessage());
			e.printStackTrace();
			Flow.currentFlow().destroy(); // XXX System.exit(1)
		}
		catch(Exception e)
		{
			Node.println("There was an exception "
				     + "in the testing code:"
				     + e.getMessage());
			e.printStackTrace();
			Flow.currentFlow().destroy(); // XXX System.exit(1)
		}
	}

	/**
	 * Overloaded method 
	 *
	 * @param name The name of the class to be tested
	 */
	public static void testClass(String name)
	{
		Class cl;

		try
		{
			cl = Class.forName(name);
			Tester.testClass(cl);
		}
		catch(ClassNotFoundException e)
		{
			Node.println("Couldn't find class "
				     + name
				     + " to test it.");
			Flow.currentFlow().destroy(); // XXX System.exit(1);
		}
	}
}
