/*
 * ForwardFlow.Java
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_TESTS_PKG.demo;

import java.io.IOException;

import JANOS_NODEOS_PKG.Flow;
import JANOS_NODEOS_PKG.Info;
import JANOS_NODEOS_PKG.Node;
import JANOS_NODEOS_PKG.FlowRun;
import JANOS_NODEOS_PKG.Address;
import JANOS_NODEOS_PKG.TimeSpec;
import JANOS_NODEOS_PKG.InChannel;
import JANOS_NODEOS_PKG.OutChannel;
import JANOS_NODEOS_PKG.DeliverFunc;
import JANOS_NODEOS_PKG.NetworkSpec;
import JANOS_NODEOS_PKG.BufferHandle;
import JANOS_NODEOS_PKG.DemultiplexKey;
import JANOS_NODEOS_PKG.ThreadPoolSpec;
import JANOS_NODEOS_PKG.CommSpaceHandle;
import JANOS_NODEOS_PKG.ThreadScheduler;
import JANOS_NODEOS_PKG.CommSpaceElement;
import JANOS_NODEOS_PKG.TransportInterface;
import JANOS_NODEOS_PKG.AddressFormatException;
import JANOS_NODEOS_PKG.ProtocolSpecFormatException;
import JANOS_NODEOS_PKG.PacketAddressFormatException;

import JANOS_PROTOCOLS_PKG.ip.IPHeader;
import JANOS_PROTOCOLS_PKG.icmp.ICMPHeader;

import JANOS_SERVICES_PKG.arp.ARPHandler;

/**
 */
class ForwardFlow
	extends SubFlow
{
	private int RATE_DIVISOR = 10;
	private int BANDWIDTH_DIVISOR = 10;
	private int MAXBUFS = 64;

	private static class RouteEntry
	{
		PACKAGE RouteEntry next;
		private final int address;
		private final OutChannel channel;

		public RouteEntry(int address, OutChannel chan)
		{
			this.address = address;
			this.channel = chan;
		}

		public RouteEntry(RouteEntry re)
		{
			this(re.address, re.channel);
		}

		public int getAddress()
		{
			return this.address;
		}

		public OutChannel getChannel()
		{
			return this.channel;
		}
		
		public int hashCode()
		{
			return this.address;
		}

		public boolean equals(int addr)
		{
			boolean retval = false;

			if( addr == this.address )
			{
				retval = true;
			}
			return retval;
		}
	}

	private static class RouteTable
	{
		private static final int TABLE_SIZE = 31;
		
		private RouteEntry entries[];

		public RouteTable()
		{
			this.entries = new RouteEntry[TABLE_SIZE];
		}

		public synchronized void add(RouteEntry re)
		{
			if( this.lookup(re.getAddress()) == null )
			{
				int hash;
				
				hash = (re.hashCode() & 0x7fffffff)
					% TABLE_SIZE;
				re.next = this.entries[hash];
				this.entries[hash] = re;
			}
		}

		public synchronized RouteEntry lookup(int address)
		{
			RouteEntry re, retval = null;
			int hash;

			hash = (address & 0x7fffffff) % TABLE_SIZE;
			re = this.entries[hash];
			while( (re != null) && (retval == null) )
			{
				if( re.equals(address) )
					retval = re;
				re = re.next;
			}
			return retval;
		}
	}
	
	private class IPReceiver
		implements DeliverFunc
	{
		private InChannel ic;
		private int iface;
		
		public IPReceiver(Flow flow, int iface,
				  String proto, String type)
			throws Exception
		{
			DemultiplexKey dk = new DemultiplexKey(flow);
			ThreadPoolSpec tps = new ThreadPoolSpec(
				1,
				32768,
				ThreadPoolSpec.getMaxRate(),
				null);
			String src;

			this.iface = iface;
			src = Info.getInterface(proto).getAddress().getSource()
				+ ":"
				+ type;
			this.ic = InChannel.create(flow,
						   tps,
						   dk,
						   proto,
						   src,
						   this);
			this.ic.addBuffers(MAXBUFS);
		}
		
		public void handlePacket(BufferHandle bh, int payloadOffset)
		{
			OutChannel oc;
			RouteEntry re;
			
			if( bh.computeChecksum(
				0, IPHeader.HEADER_LENGTH_NO_OPTIONS) != 0 )
			{
				Node.println("Bad checksum... ");
				return;
			}
			if( !IPHeader.consumeTTL(bh, 0) )
			{
				Node.println("No TTL...");
				return;
			}
			re = ForwardFlow.this.lookupRoute(
				this.iface,
				IPHeader.getDestination(bh, payloadOffset));
			if( (oc = re.getChannel()) != null )
			{
				oc.send(bh);
			}
		}

		public void receive(BufferHandle bh, int payloadOffset)
		{
			try
			{
				this.handlePacket(bh, payloadOffset);
			}
			catch(Throwable th)
			{
				Node.println("Exception during packet "
					     + "processing "
					     + th);
				Node.printStackTrace(th);
			}
			this.ic.addBuffer(bh);
		}
	}

	private RouteTable cachedRoutes[];
	private IPReceiver receivers[];

	private String getDestination(int iface, int address)
	{
		String retval;
		
		if( (retval = Flow.lookupConfig("ForwardConfig:permanent-arp",
						Address.toString(address)))
		    == null )
		{
			StringBuffer dst = new StringBuffer();
			byte ethaddr[] = { 0, 0, 0, 0, 0, 0 };
			byte ipaddr[] = { 0, 0, 0, 0 };
			int lpc;

			this.translateIP(ipaddr, address);
			ARPHandler.arpfor(iface, ipaddr, ethaddr);
			for( lpc = 0; lpc < ethaddr.length; lpc++ )
			{
				dst.append(Integer.toHexString(ethaddr[lpc] &
							       0xff));
			}
			retval = dst.toString();
		}
		Node.println("Address "
			     + Address.toString(address)
			     + " resolved to "
			     + retval);
		return retval;
	}
	
	private RouteEntry createRoute(int iface, int address, String name)
	{
		String outputName = "ForwardConfig:" + name;
		String protocolSpec, type, dst = null;
		TransportInterface ti;
		OutChannel oc = null;
		RouteEntry retval;

		protocolSpec = Flow.lookupConfig(outputName, "protocol");
		try
		{
			type = Flow.lookupConfig(outputName, "type");
			if( address != -1 )
			{
				dst = this.getDestination(iface, address);
				Node.println("Creating out channel "
					     + protocolSpec
					     + " "
					     + dst
					     + ":"
					     + type);
				oc = OutChannel.create(
					this.flow,
					new NetworkSpec(2, // XXX 50%
							NetworkSpec.DIVISOR),
					protocolSpec,
					dst + ":" + type);
			}
			else
			{
				Node.println("Ignoring broadcast packet...");
			}
		}
		catch(ProtocolSpecFormatException e)
		{
			Node.println("Error: `"
				     + protocolSpec
				     + "' Is not a valid protocol spec.");
		}
		catch(PacketAddressFormatException e)
		{
			Node.println("Error: `"
				     + dst
				     + "' Is not a valid address spec.");
		}
		catch(IOException e)
		{
			Node.println("Error: `"
				     + dst
				     + "' Is not a valid address spec.");
		}
		catch(Exception e)
		{
			Node.println("Error: `"
				     + e
				     + "' Unknown exception.");
		}
		catch(Error e)
		{
			Node.println("Error: `"
				     + e
				     + "' Unknown error.");
		}
		retval = new RouteEntry(address, oc);
		return retval;
	}

	private RouteEntry lookupRoute(int iface, int address)
	{
		RouteEntry retval;

		if( (retval = this.cachedRoutes[iface].lookup(address))
		    == null )
		{
			String str;

			// Lookup routes for this interface
			str = Flow.lookupConfig("ForwardConfig:routes" + iface,
						Address.toString(address));

			// If no routes, look for default route info
			if( str == null )
			{
				str = Flow.lookupConfig("ForwardConfig:routes"
							+ iface,
							"*");
				if( str == null )
				{
					Node.println("No default route in "
						     + "ForwardConfig:routes"
						     + iface);
				}
			}
			
			// Add the route(s)
			if( str != null )
			{
				retval = this.createRoute(iface, address, str);
			}
			else
			{
				Node.println("No route for "
					     + Address.toString(address));
				retval = new RouteEntry(address, null);
			}
			Node.println("Adding route for "
				     + Address.toString(address)
				     + " -> "
				     + retval.getChannel());
			this.cachedRoutes[iface].add(retval);
		}
		return retval;
	}
	
	private final byte[] translateIP(byte dest[], int ip)
	{
		dest[0] = (byte)((ip >> 24) & 0xFF);
		dest[1] = (byte)((ip >> 16) & 0xFF);
		dest[2] = (byte)((ip >>  8) & 0xFF);
		dest[3] = (byte)((ip      ) & 0xFF);
		return dest;
	}
	
	private final byte[] translateIP(byte dest[], String ip)
		throws AddressFormatException
	{
		return this.translateIP(dest, Address.fromString(ip));
	}
	
	private void configureARP()
	{
		byte ipaddr[] = { 0, 0, 0, 0 };
		int lpc;

		for( lpc = 0; Info.getInterface("eth" + lpc) != null; lpc++ )
		{
			String str;
			
			str = Flow.lookupConfig("ForwardConfig:arp",
						Integer.toString(lpc));
			if( str != null )
			{
				try
				{
					Node.println("Adding arp for if"
						     + lpc
						     + " -> "
						     + str);
					this.translateIP(ipaddr, str);
					ARPHandler.config_iface(lpc, ipaddr);
				}
				catch(AddressFormatException e)
				{
					Node.println("Bad arp ip address `"
						     + str
						     + "' for if"
						     + lpc);
				}
			}
		}
	}

	// controller constructor
	ForwardFlow()
	{
		super(null);
		
		try
		{
			String str;
			
			str = Flow.lookupConfig("ForwardConfig", "rate-divisor");
			if( str != null )
				RATE_DIVISOR = Integer.parseInt(str);
		}
		catch(Exception e)
		{
			Node.println("ForwardFlow configuration failed: " + e);
			Node.printStackTrace(e);
		}
	}

	private void configure()
		throws Exception
	{
		int lpc, count;
		String str;

		count = Info.getInterfaceCount();
		this.cachedRoutes = new RouteTable[count];
		for( lpc = 0; lpc < count; lpc++ )
		{
			this.cachedRoutes[lpc] = new RouteTable();
		}
		this.receivers = new IPReceiver[count];
		str = Flow.lookupConfig("ForwardConfig", "rate-divisor");
		if( str != null )
			RATE_DIVISOR = Integer.parseInt(str);
		str = Flow.lookupConfig("ForwardConfig", "band-divisor");
		if( str != null )
			BANDWIDTH_DIVISOR = Integer.parseInt(str);
		str = Flow.lookupConfig("ForwardConfig", "max-buffers");
		if( str != null )
			MAXBUFS = Integer.parseInt(str);
		this.configureARP();
		for( lpc = 0; lpc < count; lpc++ )
		{
			if( (str = Flow.lookupConfig("ForwardConfig:input"
						     + lpc,
						     "protocol")) != null )
			{
				Node.println("Adding forwarder for "
					     + str);
				this.receivers[lpc] = new IPReceiver(
					this.flow,
					lpc,
					str,
					Flow.lookupConfig("ForwardConfig:input"
							  + lpc,
							  "type"));
			}
		}
	}

	private void printConfiguration()
	{
		Node.println("ForwardFlow configuration:");
		Node.println("  Rate divisor\t" + RATE_DIVISOR);
		Node.println("  Band divisor\t" + BANDWIDTH_DIVISOR);
		Node.println("  Max buffers\t" + MAXBUFS);
	}

	
	void doRun(CommSpaceElement[] args)
	{
		try
		{
			this.configure();
			this.printConfiguration();
		}
		catch(Exception e)
		{
			Node.println("ForwardHog configuration failed: " + e);
			Node.printStackTrace(e);
		}
	}
	
	void doStartMisbehaving()
	{
		// This flow does not misbehave
		return;
	}
	
	void doStopMisbehaving()
	{
		// This flow does not misbehave
		return;
	}

	PACKAGE void kick(Flow pflow)
	{
		int rate = ThreadPoolSpec.getMaxRate() / RATE_DIVISOR;
		
		super.kick(pflow,
			   new ThreadPoolSpec(2,
					      20 * 1024,
					      rate,
					      null));
	}
}
