/*
 * CommSpaceHandleTester.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

package JANOS_TESTS_NODEOS_PKG;

import JANOS_TESTS_PKG.*;
import JANOS_NODEOS_PKG.*;

public class CommSpaceHandleTester extends Tester
{
	private final int startValue = 0;
	private final int exitValue = 10;

	private TestFailedException consumerException = null;

	/**
	 * Runnable thread that consumes data from a space
	 */
	private class Consumer implements Runnable
	{
		private CommSpaceHandle csh;
		
		public Consumer(CommSpaceHandle csh)
		{
			this.csh = csh;
		}
		
		public void run()
		{
			CommSpaceElement cse;
			boolean done = false;
			
			try
			{
				while( !done )
				{
					java.lang.Thread.sleep(10);
					cse = this.csh.waitToTake(0);
					if( cse.getInt() == exitValue )
						done = true;
				}
			}
			catch(Exception e)
			{
				CommSpaceHandleTester csht =
					CommSpaceHandleTester.this;
				
				csht.consumerException =
					new TestFailedException(csht);
			}
		}
	}
	
	public void testUnit() throws TestFailedException
	{
		final int commSpaceSize = 16;
		CommSpaceHandle csh;
		int lpc;

		this.enterSection(ENV_NONE, "Constructor");
		csh = CommSpaceHandle.newCommSpace(commSpaceSize);
		this.test(csh.size() == commSpaceSize);
		csh = new CommSpaceHandle();
		csh = CommSpaceHandle.newCommSpace(csh, commSpaceSize);
		this.test(csh.size() == commSpaceSize);

		this.enterSection(ENV_NONE, "reading/writing");
		try
		{
			for( lpc = 0; lpc < commSpaceSize; lpc++ )
			{
				this.test(csh.swap(lpc, null) == null);
				this.test(csh.read(lpc) == null);
				this.test(csh.take(lpc) == null);
			}
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			throw new TestFailedException(this);
		}

		/* Make sure we can't index something bogus */
		int badIndexes[] =
		{
			Integer.MIN_VALUE,
			-1,
			commSpaceSize,
			commSpaceSize + 1,
			Integer.MAX_VALUE /* check for rollover */
		};
		for( lpc = 0; lpc < badIndexes.length; lpc++ )
		{
			try
			{
				csh.read(badIndexes[lpc]);
				throw new TestFailedException(this);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
			try
			{
				csh.write(badIndexes[lpc], null);
				throw new TestFailedException(this);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
			try
			{
				csh.take(badIndexes[lpc]);
				throw new TestFailedException(this);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
			try
			{
				csh.waitToTake(badIndexes[lpc]);
				throw new TestFailedException(this);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
			catch(InterruptedException e)
			{
			}
			try
			{
				csh.swap(badIndexes[lpc], null);
				throw new TestFailedException(this);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
		}
		/* Write/read/swap/take/read stuff in the comm space */
		try
		{
			CommSpaceElement cse;

			for( lpc = 0; lpc < commSpaceSize; lpc++ )
			{
				csh.write(lpc,
					  new CommSpaceElement(lpc + 1));
			}
			for( lpc = 0; lpc < commSpaceSize; lpc++ )
			{
				this.test(csh.read(lpc).getInt() == lpc + 1);
				cse = new CommSpaceElement(lpc + 2);
				this.test(csh.swap(lpc, cse).getInt() ==
					  lpc + 1);
				this.test(csh.take(lpc).getInt() == lpc + 2);
				this.test(csh.read(lpc) == null);
			}
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			throw new TestFailedException(this);
		}

		java.lang.Thread consumer;

		/* Check for synchronization */
		consumer = new java.lang.Thread(new Consumer(csh));
		try
		{
			consumer.start();
			for( lpc = startValue; lpc < exitValue; lpc++ )
			{
				java.lang.Thread.sleep(50);
				csh.write(0, new CommSpaceElement(lpc));
			}
			csh.write(0, null);
			for( lpc = startValue; lpc <= exitValue; lpc++ )
			{
				java.lang.Thread.sleep(50);
				csh.swap(0, new CommSpaceElement(lpc));
			}
			consumer.join(1000);
			if( consumer.isAlive() )
			{
				consumer.interrupt();
				throw new TestFailedException(this);
			}
		}
		catch(Exception e)
		{
			consumer.interrupt();
			throw new TestFailedException(this);
		}
		if( this.consumerException != null )
			throw this.consumerException;
	}
}
