/*
 * Packet.Java
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Janos Java NodeOS software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation is hereby granted without fee, provided that the above
 * copyright notice and this permission/disclaimer notice is retained in all
 * copies or modified versions, and that both notices appear in supporting
 * documentation.  THE COPYRIGHT HOLDERS PROVIDE THIS SOFTWARE "AS IS" AND
 * WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE COPYRIGHT HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */
package JANOS_UTIL_PKG;

import JANOS_NODEOS_PKG.BufferHandle;

/**
 * Packet is a helper class that can be used to build packets from several
 * PacketSegments.
 */
public class Packet
{
	/**
	 * Array of packet segments that make up this packet.
	 */
	private PacketSegment segments[];

	/**
	 * Records the offset in the buffer that each segment will be at
	 * when marshalled.
	 */
	private int segmentOffsets[];

	/**
	 * The total length of the packet.
	 */
	private int totalLength;

	/**
	 * @param segs The array of segments that make up this Packet.
	 */
	public Packet(PacketSegment segs[])
	{
		int lpc;
		
		this.segments = segs;
		this.segmentOffsets = new int[segs.length];
		this.totalLength = 0;
		for( lpc = 0; lpc < this.segments.length; lpc++ )
		{
			this.segmentOffsets[lpc] = this.totalLength;
			this.totalLength += this.segments[lpc].getLength();
		}
	}

	/**
	 * @return The array of packet segments that make up this Packet.
	 */
	public PacketSegment[] getSegments()
	{
		return this.segments;
	}

	/**
	 * @return The total length of this packet.
	 */
	public int getLength()
	{
		return this.totalLength;
	}

	/**
	 * @param bh The buffer to marshal this packet into.
	 * @param offset The offset in the buffer to start marshalling.
	 */
	public void marshal(BufferHandle bh, int offset)
	{
		int lpc;

		this.getLength();
		/*
		 * Marshal each segment starting with the innermost.  We do
		 * this so that the header segments can compute checksums and
		 * other data dependent stuff on the body of the packet.
		 */
		for( lpc = this.segments.length - 1; lpc >= 0; lpc-- )
		{
			this.segments[lpc].marshal(bh,
						   offset
						   + this.segmentOffsets[lpc]);
		}
	}

	/**
	 * Marshal this packet into the beginning of the given buffer.
	 *
	 * @param bh The buffer to marshal this packet into.
	 */
	public void marshal(BufferHandle bh)
	{
		this.marshal(bh, 0);
	}

	/**
	 * @param bh The buffer that contains the marshalled packet segments.
	 * @param offset The offset in the buffer to start unmarshalling.
	 */
	public void unmarshal(BufferHandle bh, int offset)
		throws InvalidPacketException
	{
		int lpc, nextOffset = offset;

		/* Unmarshal each segment of the packet */
		for( lpc = 0; lpc < this.segments.length; lpc++ )
		{
			this.segmentOffsets[lpc] = nextOffset;
			nextOffset = this.segments[lpc].
				unmarshal(bh, this.segmentOffsets[lpc]);
		}
	}

	/**
	 * Unmarshal this packet from the beginning of the given buffer.
	 *
	 * @param bh The buffer that contains the marshalled packet segments.
	 */
	public void unmarshal(BufferHandle bh)
		throws InvalidPacketException
	{
		this.unmarshal(bh, 0);
	}

	public String toString()
	{
		String retval = "";
		int lpc;

		for( lpc = 0; lpc < this.segments.length; lpc++ )
		{
			retval += this.segments[lpc];
		}
		return retval;
	}
}
